# Copyright 2018 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from typing import Any

from keystone.api.validation import parameter_types
from keystone.api.validation import response_types

# Individual properties of 'Access Rule'
_access_rules_properties = {
    "path": {
        "type": "string",
        "minLength": 0,
        "maxLength": 225,
        "pattern": r'^/\.*',
        "description": (
            "The API path that the application credential is "
            "permitted to access."
        ),
    },
    "service": {
        **parameter_types.id_string,
        "description": (
            "The service type identifier for the service that the application"
            " credential is permitted to access. Must be a service type that"
            " is listed in the service catalog and not a code name for a"
            " service."
        ),
    },
    "method": {
        "type": "string",
        "enum": ["DELETE", "GET", "HEAD", "PATCH", "POST", "PUT"],
        "description": (
            "The request method that the application credential is "
            "permitted to use for a given API endpoint."
        ),
    },
}

# Common schema of the `Access Rule` resource
access_rule_schema: dict[str, Any] = {
    "type": "object",
    "description": "An access rule object.",
    "properties": {
        "id": {
            "type": "string",
            "readOnly": True,
            "description": "The UUID of the access rule",
        },
        "links": response_types.resource_links,
        **_access_rules_properties,
    },
    "additionalProperties": False,
}

# Query parameters of the `/users/{user_id}/access_rules` and
# `/application_credentials/{application_credential_id}` APIs
index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {},
    "additionalProperties": False,
}

# Response of the `/access_rules` API
rule_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "access_rules": {
            "type": "array",
            "items": access_rule_schema,
            "description": "A list of access_rule objects.",
        },
        "links": response_types.links,
    },
    "additionalProperties": False,
}

# /access/rules/{access_rule_id}
# GET request query parameters
rule_show_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {},
    "additionalProperties": False,
}

# Response of `/access_rules/{access_rule_id}` API returning
# single access rule
rule_show_response_body: dict[str, Any] = {
    "type": "object",
    "description": "An access rule object.",
    "properties": {"access_rule": access_rule_schema},
    "additionalProperties": False,
}

# Individual properties of 'Application Credential'
_application_credential_properties = {
    "name": {
        **parameter_types.name,
        "description": (
            "The name of the application credential. Must be unique to a user."
        ),
    },
    "description": {
        "type": ["string", "null"],
        "description": (
            "A description of the application credential's purpose."
        ),
    },
    "expires_at": {
        "type": ["string", "null"],
        "description": (
            "The expiration time of the application credential, if one "
            "was specified."
        ),
    },
    "project_id": {
        "type": "string",
        "description": (
            "The ID of the project the application credential was "
            "created for and that authentication requests using this "
            "application credential will be scoped to."
        ),
    },
    "access_rules": {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "id": parameter_types.id_string,
                **_access_rules_properties,
            },
        },
        "description": "A list of access_rules objects.",
    },
    "unrestricted": {
        "type": ["boolean", "null"],
        "description": (
            "A flag indicating whether the application credential "
            "may be used for creation or destruction of other "
            "application credentials or trusts."
        ),
    },
    "system": {"type": ["string", "null"]},
}

# Common schema of `Application Credential` resource
application_credential_schema: dict[str, Any] = {
    "type": "object",
    "description": "An application credential object.",
    "properties": {
        "id": {
            "type": "string",
            "readOnly": True,
            "description": "The UUID of the application credential",
        },
        "roles": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": parameter_types.id_string,
                    "name": parameter_types.name,
                    "domain_id": {
                        "type": ["string", "null"],
                        "description": "The ID of the domain of the role.",
                    },
                    "description": {
                        "type": ["string", "null"],
                        "description": "A description about the role.",
                    },
                    "options": {
                        "type": ["object", "null"],
                        "description": (
                            "The resource options for the role. "
                            "Available resource options are immutable."
                        ),
                    },
                },
                "additionalProperties": False,
            },
            "description": (
                "A list of one or more roles that this application "
                "credential has associated with its project. A token "
                "using this application credential will have these "
                "same roles."
            ),
        },
        "user_id": {"type": "string", "description": "The ID of the user."},
        "links": response_types.resource_links,
        **_application_credential_properties,
    },
    "additionalProperties": False,
}

# Query parameters of `/application_credentials` API
application_credential_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": (
                "The name of the application credential. "
                "Must be unique to a user."
            ),
        }
    },
    "additionalProperties": False,
}

# Response of the `/application_credentials` API
application_credential_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "application_credentials": {
            "type": "array",
            "items": application_credential_schema,
            "description": "A list of application credentials",
        },
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

application_credential_request_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "id": {
            "type": "string",
            "description": "The UUID of the application credential",
        }
    },
    "additionalProperties": False,
}

# Response of `/application_credentials/{application_credential_id}`
# API returning single access rule
application_credential_response_body: dict[str, Any] = {
    "type": "object",
    "description": "An application credential object.",
    "properties": {"application_credential": application_credential_schema},
    "additionalProperties": False,
}

# Request body of the `POST /application_credentials` operation
application_credential_create_request_body: dict[str, Any] = {
    "type": "object",
    "description": "An application credential object.",
    "properties": {
        "application_credential": {
            "type": "object",
            "properties": {
                "id": {
                    "type": "string",
                    "description": "The UUID for the credential.",
                },
                "secret": {
                    "type": ["string", "null"],
                    "description": (
                        "The secret that the application credential "
                        "will be created with. If not provided, one "
                        "will be generated."
                    ),
                },
                **_application_credential_properties,
                "roles": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "id": parameter_types.id_string,
                            "name": parameter_types.name,
                        },
                        "minProperties": 1,
                        "maxProperties": 1,
                        "additionalProperties": False,
                    },
                    "description": (
                        "A list of one or more roles that this application "
                        "credential has associated with its project. A token "
                        "using this application credential will have these "
                        "same roles."
                    ),
                },
            },
            "additionalProperties": False,
            "required": ["name"],
        }
    },
    "required": ["application_credential"],
    "additionalProperties": False,
}

application_credential_create_response_body = copy.deepcopy(
    application_credential_response_body
)
application_credential_create_response_body["properties"][
    "application_credential"
]["properties"]["secret"] = {
    "type": "string",
    "description": (
        "The secret that the application credential will be created with."
    ),
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import sql
from keystone import exception
from keystone.i18n import _
from keystone.resource.config_backends import base


class WhiteListedConfig(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'whitelisted_config'
    domain_id = sql.Column(sql.String(64), primary_key=True)
    group = sql.Column(sql.String(255), primary_key=True)
    option = sql.Column(sql.String(255), primary_key=True)
    value = sql.Column(sql.JsonBlob(), nullable=False)

    def to_dict(self):
        d = super().to_dict()
        d.pop('domain_id')
        return d


class SensitiveConfig(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'sensitive_config'
    domain_id = sql.Column(sql.String(64), primary_key=True)
    group = sql.Column(sql.String(255), primary_key=True)
    option = sql.Column(sql.String(255), primary_key=True)
    value = sql.Column(sql.JsonBlob(), nullable=False)

    def to_dict(self):
        d = super().to_dict()
        d.pop('domain_id')
        return d


class ConfigRegister(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'config_register'
    type = sql.Column(sql.String(64), primary_key=True)
    domain_id = sql.Column(sql.String(64), nullable=False)


class DomainConfig(base.DomainConfigDriverBase):
    def choose_table(self, sensitive):
        if sensitive:
            return SensitiveConfig
        else:
            return WhiteListedConfig

    def _create_config_option(
        self, session, domain_id, group, option, sensitive, value
    ):
        config_table = self.choose_table(sensitive)
        ref = config_table(
            domain_id=domain_id, group=group, option=option, value=value
        )
        session.add(ref)

    def create_config_options(self, domain_id, option_list):
        with sql.session_for_write() as session:
            for config_table in [WhiteListedConfig, SensitiveConfig]:
                query = session.query(config_table)
                query = query.filter_by(domain_id=domain_id)
                query.delete(False)
            for option in option_list:
                self._create_config_option(
                    session,
                    domain_id,
                    option['group'],
                    option['option'],
                    option['sensitive'],
                    option['value'],
                )

    def _get_config_option(self, session, domain_id, group, option, sensitive):
        try:
            config_table = self.choose_table(sensitive)
            ref = (
                session.query(config_table)
                .filter_by(domain_id=domain_id, group=group, option=option)
                .one()
            )
        except sql.NotFound:
            msg = _('option %(option)s in group %(group)s') % {
                'group': group,
                'option': option,
            }
            raise exception.DomainConfigNotFound(
                domain_id=domain_id, group_or_option=msg
            )
        return ref

    def get_config_option(self, domain_id, group, option, sensitive=False):
        with sql.session_for_read() as session:
            ref = self._get_config_option(
                session, domain_id, group, option, sensitive
            )
            return ref.to_dict()

    def list_config_options(
        self, domain_id, group=None, option=None, sensitive=False
    ):
        with sql.session_for_read() as session:
            config_table = self.choose_table(sensitive)
            query = session.query(config_table)
            query = query.filter_by(domain_id=domain_id)
            if group:
                query = query.filter_by(group=group)
                if option:
                    query = query.filter_by(option=option)
            return [ref.to_dict() for ref in query.all()]

    def update_config_options(self, domain_id, option_list):
        with sql.session_for_write() as session:
            for option in option_list:
                self._delete_config_options(
                    session, domain_id, option['group'], option['option']
                )
                self._create_config_option(
                    session,
                    domain_id,
                    option['group'],
                    option['option'],
                    option['sensitive'],
                    option['value'],
                )

    def _delete_config_options(self, session, domain_id, group, option):
        for config_table in [WhiteListedConfig, SensitiveConfig]:
            query = session.query(config_table)
            query = query.filter_by(domain_id=domain_id)
            if group:
                query = query.filter_by(group=group)
                if option:
                    query = query.filter_by(option=option)
            query.delete(False)

    def delete_config_options(self, domain_id, group=None, option=None):
        with sql.session_for_write() as session:
            self._delete_config_options(session, domain_id, group, option)

    def obtain_registration(self, domain_id, type):
        try:
            with sql.session_for_write() as session:
                ref = ConfigRegister(type=type, domain_id=domain_id)
                session.add(ref)
            return True
        except sql.DBDuplicateEntry:  # nosec
            # Continue on and return False to indicate failure.
            pass
        return False

    def read_registration(self, type):
        with sql.session_for_read() as session:
            ref = session.get(ConfigRegister, type)
            if not ref:
                raise exception.ConfigRegistrationNotFound()
            return ref.domain_id

    def release_registration(self, domain_id, type=None):
        """Silently delete anything registered for the domain specified."""
        with sql.session_for_write() as session:
            query = session.query(ConfigRegister)
            if type:
                query = query.filter_by(type=type)
            query = query.filter_by(domain_id=domain_id)
            query.delete(False)



# Copyright 2013 OpenStack Foundation
# Copyright 2013 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import functools
import typing as ty

import keystone.conf
from keystone import exception
from keystone.i18n import _

CONF = keystone.conf.CONF


def truncated(f):
    """Ensure list truncation is detected in Driver list entity methods.

    This is designed to wrap Driver list_{entity} methods in order to
    calculate if the resultant list has been truncated. Provided a limit dict
    is found in the hints list, we increment the limit by one so as to ask the
    wrapped function for one more entity than the limit, and then once the list
    has been generated, we check to see if the original limit has been
    exceeded, in which case we truncate back to that limit and set the
    'truncated' boolean to 'true' in the hints limit dict.

    """

    @functools.wraps(f)
    def wrapper(self, hints, *args, **kwargs):
        if not hasattr(hints, 'limit'):
            raise exception.UnexpectedError(
                _(
                    'Cannot truncate a driver call without hints list as '
                    'first parameter after self '
                )
            )

        if hints.limit is None or hints.filters:
            return f(self, hints, *args, **kwargs)

        # A limit is set, so ask for one more entry than we need
        list_limit = hints.limit['limit']
        hints.set_limit(list_limit + 1)
        ref_list = f(self, hints, *args, **kwargs)

        # If we got more than the original limit then trim back the list and
        # mark it truncated. In both cases, make sure we set the limit back
        # to its original value.
        if len(ref_list) > list_limit:
            hints.set_limit(list_limit, truncated=True)
            return ref_list[:list_limit]
        else:
            hints.set_limit(list_limit)
            return ref_list

    return wrapper


class Hints:
    """Encapsulate driver hints for listing entities.

    Hints are modifiers that affect the return of entities from a
    list_<entities> operation.  They are typically passed to a driver to give
    direction as to what filtering, pagination or list limiting actions are
    being requested.

    It is optional for a driver to action some or all of the list hints,
    but any filters that it does satisfy must be marked as such by calling
    removing the filter from the list.

    A Hint object contains filters, which is a list of dicts that can be
    accessed publicly. Also it contains a dict called limit, which will
    indicate the amount of data we want to limit our listing to.

    If the filter is discovered to never match, then `cannot_match` can be set
    to indicate that there will not be any matches and the backend work can be
    short-circuited.

    Each filter term consists of:

    * ``name``: the name of the attribute being matched
    * ``value``: the value against which it is being matched
    * ``comparator``: the operation, which can be one of ``equals``,
                      ``contains``, ``startswith`` or ``endswith``
    * ``case_sensitive``: whether any comparison should take account of
                          case

    """

    def __init__(self):
        self.limit: ty.Optional[dict[str, ty.Any]] = None
        self.marker: ty.Optional[str] = None
        self.filters = []
        self.cannot_match = False

    def add_filter(
        self,
        name,
        value,
        comparator: str = 'equals',
        case_sensitive: bool = False,
    ):
        """Add a filter to the filters list, which is publicly accessible."""
        self.filters.append(
            {
                'name': name,
                'value': value,
                'comparator': comparator,
                'case_sensitive': case_sensitive,
            }
        )

    def get_exact_filter_by_name(self, name):
        """Return a filter key and value if exact filter exists for name."""
        for entry in self.filters:
            if entry['name'] == name and entry['comparator'] == 'equals':
                return entry

    def set_limit(self, limit: int, truncated: bool = False):
        """Set a limit to indicate the list should be truncated."""
        self.limit = {'limit': limit, 'truncated': truncated}

    def get_limit_or_max(self) -> int:
        """Get page limit or max page size

        Return page limit (size) as requested by user (or API flow) or the
        maximum page size if not present. This method is invoked by the SQL
        drivers.

        :returns int: Page size
        """
        limit: ty.Optional[int] = (
            self.limit.get("limit") if self.limit else None
        )
        return int(limit) if limit else CONF.max_db_limit

    def get_limit_with_default(
        self, default_limit: ty.Optional[int] = None
    ) -> int:
        """Return page limit for the query.

        1. `limit` was set in the query parameters:
            `min(limit, MAX_LIMIT)`

        2. `limit` is not set and `default_limit` is set:
            `min(default_limit, MAX_LIMIT)`

        2. `limit` is null, `default_limit` is null, `CONF.list_limit` is set:
            `min(CONF.list_limit, MAX_LIMIT)`

        3. `limit` is null, `default_limit` is null, `CONF.list_limit` is null:
            `CONF.max_db_limit`

        """

        _user_limit: ty.Optional[int] = (
            self.limit.get("limit") if self.limit else None
        )
        _absolute_limit: int = CONF.max_db_limit
        # when default (as resource specific) is not set try to get the global
        # default
        _default_limit: ty.Optional[int] = default_limit or CONF.list_limit
        if _user_limit:
            return min(_user_limit, _absolute_limit)
        elif _default_limit:
            return min(_default_limit, _absolute_limit)
        else:
            return CONF.max_db_limit

    def set_marker(self, marker: str):
        """Set a marker pointing to the last entry of the page."""
        self.marker = marker



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import sys

import flask
import oslo_i18n
from oslo_log import log
from oslo_middleware import healthcheck

try:
    # werkzeug 0.15.x
    from werkzeug.middleware import dispatcher as wsgi_dispatcher
except ImportError:
    # werkzeug 0.14.x
    import werkzeug.wsgi as wsgi_dispatcher

from keystone.api import auth
from keystone.api import credentials
from keystone.api import discovery
from keystone.api import domains
from keystone.api import ec2tokens
from keystone.api import endpoints
from keystone.api import groups
from keystone.api import limits
from keystone.api import os_ep_filter
from keystone.api import os_federation
from keystone.api import os_inherit
from keystone.api import os_oauth1
from keystone.api import os_oauth2
from keystone.api import os_revoke
from keystone.api import os_simple_cert
from keystone.api import policy
from keystone.api import projects
from keystone.api import regions
from keystone.api import registered_limits
from keystone.api import role_assignments
from keystone.api import role_inferences
from keystone.api import roles
from keystone.api import s3tokens
from keystone.api import services
from keystone.api import system
from keystone.api import trusts
from keystone.api import users
from keystone import exception
from keystone.oauth2 import handlers as oauth2_handlers
from keystone.receipt import handlers as receipt_handlers
from keystone.server.flask import common as ks_flask
from keystone.server.flask.request_processing import json_body
from keystone.server.flask.request_processing import req_logging

LOG = log.getLogger(__name__)

__apis__ = (
    discovery,
    auth,
    credentials,
    domains,
    ec2tokens,
    endpoints,
    groups,
    limits,
    os_ep_filter,
    os_federation,
    os_inherit,
    os_oauth1,
    os_oauth2,
    os_revoke,
    os_simple_cert,
    policy,
    projects,
    regions,
    registered_limits,
    role_assignments,
    role_inferences,
    roles,
    s3tokens,
    services,
    system,
    trusts,
    users,
)


def fail_gracefully(f):
    """Log exceptions and aborts."""

    @functools.wraps(f)
    def wrapper(*args, **kw):
        try:
            return f(*args, **kw)
        except Exception as e:
            LOG.debug(e, exc_info=True)

            # exception message is printed to all logs
            LOG.critical(e)
            sys.exit(1)

    return wrapper


def _add_vary_x_auth_token_header(response):
    # Add the expected Vary Header, this is run after every request in the
    # response-phase
    response.headers['Vary'] = 'X-Auth-Token'
    return response


def _best_match_language():
    """Determine the best available locale.

    This returns best available locale based on the Accept-Language HTTP
    header passed in the request.
    """
    if not flask.request.accept_languages:
        return None
    return flask.request.accept_languages.best_match(
        oslo_i18n.get_available_languages('keystone')
    )


def _handle_keystone_exception(error):
    # TODO(adriant): register this with its own specific handler:
    if isinstance(error, exception.InsufficientAuthMethods):
        return receipt_handlers.build_receipt(error)
    elif isinstance(error, exception.OAuth2Error):
        return oauth2_handlers.build_response(error)

    if isinstance(error, exception.RedirectRequired):
        return flask.redirect(error.redirect_url)

    # Handle logging
    if isinstance(error, exception.Unauthorized):
        LOG.warning(
            "Authorization failed. %(exception)s from %(remote_addr)s",
            {'exception': error, 'remote_addr': flask.request.remote_addr},
        )
    # NOTE(frickler): NotFound exceptions happen regularly during
    # normal operations, e.g. when doing "openstack role show member",
    # so don't make a fuss about them
    elif isinstance(error, exception.NotFound):
        LOG.debug(str(error))
    else:
        LOG.exception(str(error))

    # Render the exception to something user "friendly"
    error_message = error.args[0]
    message = oslo_i18n.translate(error_message, _best_match_language())
    if message is error_message:
        # translate() didn't do anything because it wasn't a Message,
        # convert to a string.
        message = str(message)

    body = {
        'error': {'code': error.code, 'title': error.title, 'message': message}
    }

    if isinstance(error, exception.AuthPluginException):
        body['error']['identity'] = error.authentication

    # Create the response and set status code.
    response = flask.jsonify(body)
    response.status_code = error.code

    # Add the appropriate WWW-Authenticate header for Unauthorized
    if isinstance(error, exception.Unauthorized):
        url = ks_flask.base_url()
        response.headers['WWW-Authenticate'] = f'Keystone uri="{url}"'
    return response


def _handle_unknown_keystone_exception(error):
    # translate a python exception to something we can properly render as
    # an API error.
    if isinstance(error, TypeError):
        new_exc = exception.ValidationError(error)
    else:
        new_exc = exception.UnexpectedError(error)
    return _handle_keystone_exception(new_exc)


@fail_gracefully
def application_factory(name='public'):
    if name not in ('admin', 'public'):
        raise RuntimeError(
            'Application name (for base_url lookup) must be '
            'either `admin` or `public`.'
        )

    app = flask.Flask(name)

    # Register Error Handler Function for Keystone Errors.
    # NOTE(morgan): Flask passes errors to an error handling function. All of
    # keystone's api errors are explicitly registered in
    # keystone.exception.KEYSTONE_API_EXCEPTIONS and those are in turn
    # registered here to ensure a proper error is bubbled up to the end user
    # instead of a 500 error.
    for exc in exception.KEYSTONE_API_EXCEPTIONS:
        app.register_error_handler(exc, _handle_keystone_exception)

    # Register extra (python) exceptions with the proper exception handler,
    # specifically TypeError. It will render as a 400 error, but presented in
    # a "web-ified" manner
    app.register_error_handler(TypeError, _handle_unknown_keystone_exception)

    # Add core before request functions
    app.before_request(req_logging.log_request_info)
    app.before_request(json_body.json_body_before_request)

    # Add core after request functions
    app.after_request(_add_vary_x_auth_token_header)

    # NOTE(morgan): Configure the Flask Environment for our needs.
    app.config.update(
        # We want to bubble up Flask Exceptions (for now)
        PROPAGATE_EXCEPTIONS=True
    )

    for api in __apis__:
        for api_bp in api.APIs:
            api_bp.instantiate_and_register_to_app(app)

    # Load in Healthcheck and map it to /healthcheck
    hc_app = healthcheck.Healthcheck.app_factory(
        {}, oslo_config_project='keystone'
    )

    # Use the simple form of the dispatch middleware, no extra logic needed
    # for legacy dispatching. This is to mount /healthcheck at a consistent
    # place
    app.wsgi_app = wsgi_dispatcher.DispatcherMiddleware(
        app.wsgi_app, {'/healthcheck': hc_app}
    )
    return app



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from keystone import exception


class DomainConfigDriverBase(metaclass=abc.ABCMeta):
    """Interface description for a Domain Config driver."""

    @abc.abstractmethod
    def create_config_options(self, domain_id, option_list):
        """Create config options for a domain.

        Any existing config options will first be deleted.

        :param domain_id: the domain for this option
        :param option_list: a list of dicts, each one specifying an option

        Option schema::

            type: dict
            properties:
                group:
                    type: string
                option:
                    type: string
                value:
                    type: depends on the option
                sensitive:
                    type: boolean
            required: [group, option, value, sensitive]
            additionalProperties: false

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def get_config_option(self, domain_id, group, option, sensitive=False):
        """Get the config option for a domain.

        :param domain_id: the domain for this option
        :param group: the group name
        :param option: the option name
        :param sensitive: whether the option is sensitive

        :returns: dict containing group, option and value
        :raises keystone.exception.DomainConfigNotFound: the option doesn't
                                                         exist.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def list_config_options(
        self, domain_id, group=None, option=False, sensitive=False
    ):
        """Get a config options for a domain.

        :param domain_id: the domain for this option
        :param group: optional group option name
        :param option: optional option name. If group is None, then this
                       parameter is ignored
        :param sensitive: whether the option is sensitive

        :returns: list of dicts containing group, option and value

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def update_config_options(self, domain_id, option_list):
        """Update config options for a domain.

        :param domain_id: the domain for this option
        :param option_list: a list of dicts, each one specifying an option

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def delete_config_options(self, domain_id, group=None, option=None):
        """Delete config options for a domain.

        Allows deletion of all options for a domain, all options in a group
        or a specific option. The driver is silent if there are no options
        to delete.

        :param domain_id: the domain for this option
        :param group: optional group option name
        :param option: optional option name. If group is None, then this
                       parameter is ignored

        The option is uniquely defined by domain_id, group and option,
        irrespective of whether it is sensitive ot not.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def obtain_registration(self, domain_id, type):
        """Try and register this domain to use the type specified.

        :param domain_id: the domain required
        :param type: type of registration
        :returns: True if the domain was registered, False otherwise. Failing
                  to register means that someone already has it (which could
                  even be the domain being requested).

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def read_registration(self, type):
        """Get the domain ID of who is registered to use this type.

        :param type: type of registration
        :returns: domain_id of who is registered.
        :raises keystone.exception.ConfigRegistrationNotFound: If nobody is
            registered.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def release_registration(self, domain_id, type=None):
        """Release registration if it is held by the domain specified.

        If the specified domain is registered for this domain then free it,
        if it is not then do nothing - no exception is raised.

        :param domain_id: the domain in question
        :param type: type of registration, if None then all registrations
                     for this domain will be freed

        """
        raise exception.NotImplemented()  # pragma: no cover



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Implement the "Immutable" resource option
from keystone.common.resource_options import core as ro_core
from keystone.common.validation import parameter_types
from keystone import exception

IMMUTABLE_OPT = ro_core.ResourceOption(
    option_id='IMMU',
    option_name='immutable',
    validator=ro_core.boolean_validator,
    json_schema_validation=parameter_types.boolean,
)


def check_resource_immutable(resource_ref):
    """Check to see if a resource is immutable.

    :param resource_ref: a dict reference of a resource to inspect
    """
    return resource_ref.get('options', {}).get(
        IMMUTABLE_OPT.option_name, False
    )


def check_immutable_update(
    original_resource_ref, new_resource_ref, type, resource_id
):
    """Check if an update is allowed to an immutable resource.

    Valid cases where an update is allowed:

        * Resource is not immutable
        * Resource is immutable, and update to set immutable to False or None

    :param original_resource_ref: a dict resource reference representing
                                  the current resource
    :param new_resource_ref: a dict reference of the updates to perform
    :param type: the resource type, e.g. 'project'
    :param resource_id: the id of the resource (e.g. project['id']),
                        usually a UUID
    :raises: ResourceUpdateForbidden
    """
    immutable = check_resource_immutable(original_resource_ref)
    if immutable:
        new_options = new_resource_ref.get('options', {})
        if type == "domain":
            if (
                new_resource_ref.get("is_domain", False) == True
                and not new_resource_ref.get("domain_id")
                and not new_resource_ref.get("parent_id")
            ):
                # To keep next check happy - reject certain props for the domain set by default in
                # `get_project_from_domain` if those ARE default
                new_resource_ref.pop("is_domain")
                new_resource_ref.pop("domain_id")
                new_resource_ref.pop("parent_id")
        # If resource is currently immutable - raise error in attempt to
        # update more then 1 property while making resource mutable
        # (first make mutable then update rest)
        if (
            (len(new_resource_ref.keys()) > 1)
            or (IMMUTABLE_OPT.option_name not in new_options)
            or (new_options[IMMUTABLE_OPT.option_name] not in (False, None))
        ):
            raise exception.ResourceUpdateForbidden(
                type=type, resource_id=resource_id
            )


def check_immutable_delete(resource_ref, resource_type, resource_id):
    """Check if a delete is allowed on a resource.

    :param resource_ref: dict reference of the resource
    :param resource_type: resource type (str) e.g. 'project'
    :param resource_id: id of the resource (str) e.g. project['id']
    :raises: ResourceDeleteForbidden
    """
    if check_resource_immutable(resource_ref):
        raise exception.ResourceDeleteForbidden(
            type=resource_type, resource_id=resource_id
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Remove duplicate constraints.

Revision ID: c88cdce8f248
Revises: 99de3849d860
Create Date: 2023-03-15 13:17:44.060715
"""

from alembic import op
from sqlalchemy.engine import reflection

# revision identifiers, used by Alembic.
revision = 'c88cdce8f248'
down_revision = '99de3849d860'
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_bind()

    # This only affects MySQL - PostgreSQL and SQLite were smart enough to
    # ignore the duplicate constraints
    if bind.engine.name != 'mysql':
        return

    # We want to drop a duplicate index on the 'project_tag' table. To drop an
    # index, we would normally just use drop_index like so:
    #
    #   with op.batch_alter_table('project_tag', schema=None) as batch_op:
    #       batch_op.drop_index(foo)
    #
    # However, the index wasn't explicitly named so we're not sure what 'foo'
    # is. It has two potential names:
    #
    # - If it was created by alembic, alembic will have left things up to the
    #   backend, which on MySQL means the index will have the same name as the
    #   first column the index covers [1]. Alternatively, ...
    # - If it was created by sqlalchemy-migrate then it will be called
    #   '{table_name}_{first_column_name}_key' [2]
    #
    # We need to handle both, so we need to first inspect the table to find
    # which it is.
    #
    # Note that unlike MariaDB [3], MySQL [4] doesn't support the 'ALTER TABLE
    # tbl_name DROP CONSTRAINT IF EXISTS constraint_name' syntax, which would
    # have allowed us to avoid the inspection. Boo.
    #
    # [1] https://dba.stackexchange.com/a/160712
    # [2] https://opendev.org/x/sqlalchemy-migrate/src/commit/5d1f322542cd8eb42381612765be4ed9ca8105ec/migrate/changeset/constraint.py#L199  # noqa: E501
    # [3] https://mariadb.com/kb/en/alter-table/
    # [4] https://dev.mysql.com/doc/refman/8.0/en/alter-table.html

    inspector = reflection.Inspector.from_engine(bind)
    indexes = inspector.get_indexes('project_tag')

    index_name = None
    for index in indexes:
        if index['column_names'] == ['project_id', 'name']:
            index_name = index['name']
            break
    else:
        # This should never happen *but* we silently ignore it since there's no
        # need to break user's upgrade flow, even if they've borked something
        return

    with op.batch_alter_table('project_tag', schema=None) as batch_op:
        batch_op.drop_index(index_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client

# This file handles all flask-restful resources for /v3/auth
import string  # noqa: I202
import urllib

import flask
import flask_restful
from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import strutils
import werkzeug.exceptions

from keystone.api._shared import authentication
from keystone.api._shared import json_home_relations
from keystone.api._shared import saml
from keystone.auth import schema as auth_schema
from keystone.common import authorization
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone.common import render_token
from keystone.common import utils as k_utils
from keystone.common import validation
import keystone.conf
from keystone import exception
from keystone.federation import idp as keystone_idp
from keystone.federation import schema as federation_schema
from keystone.federation import utils as federation_utils
from keystone.i18n import _
from keystone.server import flask as ks_flask

CONF = keystone.conf.CONF
ENFORCER = rbac_enforcer.RBACEnforcer
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs


def _combine_lists_uniquely(a, b):
    # it's most likely that only one of these will be filled so avoid
    # the combination if possible.
    if a and b:
        return {x['id']: x for x in a + b}.values()
    else:
        return a or b


def _build_response_headers(service_provider):
    # URLs in header are encoded into bytes
    return [
        ('Content-Type', 'text/xml'),
        ('X-sp-url', service_provider['sp_url'].encode('utf-8')),
        ('X-auth-url', service_provider['auth_url'].encode('utf-8')),
    ]


def _get_sso_origin_host():
    """Validate and return originating dashboard URL.

    Make sure the parameter is specified in the request's URL as well its
    value belongs to a list of trusted dashboards.

    :raises keystone.exception.ValidationError: ``origin`` query parameter
        was not specified. The URL is deemed invalid.
    :raises keystone.exception.Unauthorized: URL specified in origin query
        parameter does not exist in list of websso trusted dashboards.
    :returns: URL with the originating dashboard

    """
    origin = flask.request.args.get('origin')

    if not origin:
        msg = 'Request must have an origin query parameter'
        tr_msg = _('Request must have an origin query parameter')
        LOG.error(msg)
        raise exception.ValidationError(tr_msg)

    host = urllib.parse.unquote_plus(origin)

    # change trusted_dashboard hostnames to lowercase before comparison
    trusted_dashboards = [
        k_utils.lower_case_hostname(trusted)
        for trusted in CONF.federation.trusted_dashboard
    ]

    if host not in trusted_dashboards:
        msg = f'{host} is not a trusted dashboard host'
        tr_msg = _('%(host)s is not a trusted dashboard host') % {'host': host}
        LOG.error(msg)
        raise exception.Unauthorized(tr_msg)

    return host


class _AuthFederationWebSSOBase(ks_flask.ResourceBase):
    @staticmethod
    def _render_template_response(host, token_id):
        with open(CONF.federation.sso_callback_template) as template:
            src = string.Template(template.read())
        subs = {'host': host, 'token': token_id}
        body = src.substitute(subs)
        resp = flask.make_response(body, http.client.OK)
        resp.charset = 'utf-8'
        resp.headers['Content-Type'] = 'text/html'
        return resp


class AuthProjectsResource(ks_flask.ResourceBase):
    collection_key = 'projects'
    member_key = 'project'

    def get(self):
        """Get possible project scopes for token.

        GET/HEAD /v3/auth/projects
        GET/HEAD /v3/OS-FEDERATION/projects
        """
        ENFORCER.enforce_call(action='identity:get_auth_projects')
        user_id = self.auth_context.get('user_id')
        group_ids = self.auth_context.get('group_ids')

        user_p_refs = []
        grp_p_refs = []

        if user_id:
            try:
                user_p_refs = PROVIDERS.assignment_api.list_projects_for_user(
                    user_id
                )
            except exception.UserNotFound:  # nosec
                # federated users have an id but they don't link to anything
                pass

        if group_ids:
            grp_p_refs = PROVIDERS.assignment_api.list_projects_for_groups(
                group_ids
            )
        refs = _combine_lists_uniquely(user_p_refs, grp_p_refs)
        return self.wrap_collection(refs)


class AuthDomainsResource(ks_flask.ResourceBase):
    collection_key = 'domains'
    member_key = 'domain'

    def get(self):
        """Get possible domain scopes for token.

        GET/HEAD /v3/auth/domains
        GET/HEAD /v3/OS-FEDERATION/domains
        """
        ENFORCER.enforce_call(action='identity:get_auth_domains')
        user_id = self.auth_context.get('user_id')
        group_ids = self.auth_context.get('group_ids')

        user_d_refs = []
        grp_d_refs = []

        if user_id:
            try:
                user_d_refs = PROVIDERS.assignment_api.list_domains_for_user(
                    user_id
                )
            except exception.UserNotFound:  # nosec
                # federated users have an id but they don't link to anything
                pass

        if group_ids:
            grp_d_refs = PROVIDERS.assignment_api.list_domains_for_groups(
                group_ids
            )

        refs = _combine_lists_uniquely(user_d_refs, grp_d_refs)
        return self.wrap_collection(refs)


class AuthSystemResource(_AuthFederationWebSSOBase):
    def get(self):
        """Get possible system scopes for token.

        GET/HEAD /v3/auth/system
        """
        ENFORCER.enforce_call(action='identity:get_auth_system')
        user_id = self.auth_context.get('user_id')
        group_ids = self.auth_context.get('group_ids')

        user_assignments = []
        group_assignments = []

        if user_id:
            try:
                user_assignments = (
                    PROVIDERS.assignment_api.list_system_grants_for_user(
                        user_id
                    )
                )
            except exception.UserNotFound:  # nosec
                # federated users have an id but they don't link to anything
                pass

        if group_ids:
            group_assignments = (
                PROVIDERS.assignment_api.list_system_grants_for_groups(
                    group_ids
                )
            )

        assignments = _combine_lists_uniquely(
            user_assignments, group_assignments
        )

        if assignments:
            response = {
                'system': [{'all': True}],
                'links': {'self': ks_flask.base_url(path='auth/system')},
            }
        else:
            response = {
                'system': [],
                'links': {'self': ks_flask.base_url(path='auth/system')},
            }
        return response


class AuthCatalogResource(_AuthFederationWebSSOBase):
    def get(self):
        """Get service catalog for token.

        GET/HEAD /v3/auth/catalog
        """
        ENFORCER.enforce_call(action='identity:get_auth_catalog')
        user_id = self.auth_context.get('user_id')
        project_id = self.auth_context.get('project_id')

        if not project_id:
            raise exception.Forbidden(
                _(
                    'A project-scoped token is required to produce a '
                    'service catalog.'
                )
            )

        return {
            'catalog': PROVIDERS.catalog_api.get_v3_catalog(
                user_id, project_id
            ),
            'links': {'self': ks_flask.base_url(path='auth/catalog')},
        }


class AuthTokenOSPKIResource(flask_restful.Resource):
    @ks_flask.unenforced_api
    def get(self):
        """Deprecated; get revoked token list.

        GET/HEAD /v3/auth/tokens/OS-PKI/revoked
        """
        if not CONF.token.revoke_by_id:
            raise exception.Gone()
        # NOTE(lbragstad): This API is deprecated and isn't supported. Keystone
        # also doesn't store tokens, so returning a list of revoked tokens
        # would require keystone to write invalid tokens to disk, which defeats
        # the purpose. Return a 403 instead of removing the API altogether.
        raise exception.Forbidden()


class AuthTokenResource(_AuthFederationWebSSOBase):
    def get(self):
        """Validate a token.

        HEAD/GET /v3/auth/tokens
        """
        # TODO(morgan): eliminate the check_token action only use validate
        # NOTE(morgan): Well lookie here, we have different enforcements
        # for no good reason (historical), because the methods previously
        # had to be named different names. Check which method and do the
        # correct enforcement.
        if flask.request.method == 'HEAD':
            ENFORCER.enforce_call(action='identity:check_token')
        else:
            ENFORCER.enforce_call(action='identity:validate_token')

        token_id = flask.request.headers.get(
            authorization.SUBJECT_TOKEN_HEADER
        )
        access_rules_support = flask.request.headers.get(
            authorization.ACCESS_RULES_HEADER
        )
        allow_expired = strutils.bool_from_string(
            flask.request.args.get('allow_expired')
        )
        window_secs = CONF.token.allow_expired_window if allow_expired else 0
        include_catalog = 'nocatalog' not in flask.request.args
        token = PROVIDERS.token_provider_api.validate_token(
            token_id,
            window_seconds=window_secs,
            access_rules_support=access_rules_support,
        )
        token_resp = render_token.render_token_response_from_model(
            token, include_catalog=include_catalog
        )
        resp_body = jsonutils.dumps(token_resp)
        response = flask.make_response(resp_body, http.client.OK)
        response.headers['X-Subject-Token'] = token_id
        response.headers['Content-Type'] = 'application/json'
        return response

    @ks_flask.unenforced_api
    def post(self):
        """Issue a token.

        POST /v3/auth/tokens
        """
        include_catalog = 'nocatalog' not in flask.request.args
        auth_data = self.request_body_json.get('auth')
        auth_schema.validate_issue_token_auth(auth_data)
        token = authentication.authenticate_for_token(auth_data)
        resp_data = render_token.render_token_response_from_model(
            token, include_catalog=include_catalog
        )
        resp_body = jsonutils.dumps(resp_data)
        response = flask.make_response(resp_body, http.client.CREATED)
        response.headers['X-Subject-Token'] = token.id
        response.headers['Content-Type'] = 'application/json'
        return response

    def delete(self):
        """Revoke a token.

        DELETE /v3/auth/tokens
        """
        ENFORCER.enforce_call(action='identity:revoke_token')
        token_id = flask.request.headers.get(
            authorization.SUBJECT_TOKEN_HEADER
        )
        PROVIDERS.token_provider_api.revoke_token(token_id)
        return None, http.client.NO_CONTENT


class AuthFederationWebSSOResource(_AuthFederationWebSSOBase):
    @classmethod
    def _perform_auth(cls, protocol_id):
        idps = PROVIDERS.federation_api.list_idps()
        remote_id = None
        for idp in idps:
            try:
                remote_id_name = federation_utils.get_remote_id_parameter(
                    idp, protocol_id
                )
            except exception.FederatedProtocolNotFound:
                # no protocol for this IdP, so this can't be the IdP we're
                # looking for
                continue
            remote_id = flask.request.environ.get(remote_id_name)
            if remote_id:
                break
        if not remote_id:
            msg = 'Missing entity ID from environment'
            tr_msg = _('Missing entity ID from environment')
            LOG.error(msg)
            raise exception.Unauthorized(tr_msg)

        host = _get_sso_origin_host()
        ref = PROVIDERS.federation_api.get_idp_from_remote_id(remote_id)
        identity_provider = ref['idp_id']
        token = authentication.federated_authenticate_for_token(
            identity_provider=identity_provider, protocol_id=protocol_id
        )
        return cls._render_template_response(host, token.id)

    @ks_flask.unenforced_api
    def get(self, protocol_id):
        return self._perform_auth(protocol_id)

    @ks_flask.unenforced_api
    def post(self, protocol_id):
        return self._perform_auth(protocol_id)


class AuthFederationWebSSOIDPsResource(_AuthFederationWebSSOBase):
    @classmethod
    def _perform_auth(cls, idp_id, protocol_id):
        host = _get_sso_origin_host()

        token = authentication.federated_authenticate_for_token(
            identity_provider=idp_id, protocol_id=protocol_id
        )
        return cls._render_template_response(host, token.id)

    @ks_flask.unenforced_api
    def get(self, idp_id, protocol_id):
        return self._perform_auth(idp_id, protocol_id)

    @ks_flask.unenforced_api
    def post(self, idp_id, protocol_id):
        return self._perform_auth(idp_id, protocol_id)


class AuthFederationSaml2Resource(_AuthFederationWebSSOBase):
    def get(self):
        raise werkzeug.exceptions.MethodNotAllowed(valid_methods=['POST'])

    @ks_flask.unenforced_api
    def post(self):
        """Exchange a scoped token for a SAML assertion.

        POST /v3/auth/OS-FEDERATION/saml2
        """
        auth = self.request_body_json.get('auth')
        validation.lazy_validate(federation_schema.saml_create, auth)
        response, service_provider = saml.create_base_saml_assertion(auth)
        headers = _build_response_headers(service_provider)
        response = flask.make_response(response.to_string(), http.client.OK)
        for header, value in headers:
            response.headers[header] = value
        return response


class AuthFederationSaml2ECPResource(_AuthFederationWebSSOBase):
    def get(self):
        raise werkzeug.exceptions.MethodNotAllowed(valid_methods=['POST'])

    @ks_flask.unenforced_api
    def post(self):
        """Exchange a scoped token for an ECP assertion.

        POST /v3/auth/OS-FEDERATION/saml2/ecp
        """
        auth = self.request_body_json.get('auth')
        validation.lazy_validate(federation_schema.saml_create, auth)
        saml_assertion, service_provider = saml.create_base_saml_assertion(
            auth
        )
        relay_state_prefix = service_provider['relay_state_prefix']

        generator = keystone_idp.ECPGenerator()
        ecp_assertion = generator.generate_ecp(
            saml_assertion, relay_state_prefix
        )
        headers = _build_response_headers(service_provider)
        response = flask.make_response(
            ecp_assertion.to_string(), http.client.OK
        )
        for header, value in headers:
            response.headers[header] = value
        return response


class AuthAPI(ks_flask.APIBase):
    _name = 'auth'
    _import_name = __name__
    resources = []
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=AuthProjectsResource,
            url='/auth/projects',
            alternate_urls=[
                {
                    'url': '/OS-FEDERATION/projects',
                    'json_home': ks_flask.construct_json_home_data(
                        rel='projects',
                        resource_relation_func=(
                            json_home_relations.os_federation_resource_rel_func
                        ),
                    ),
                }
            ],
            rel='auth_projects',
            resource_kwargs={},
        ),
        ks_flask.construct_resource_map(
            resource=AuthDomainsResource,
            url='/auth/domains',
            alternate_urls=[
                {
                    'url': '/OS-FEDERATION/domains',
                    'json_home': ks_flask.construct_json_home_data(
                        rel='domains',
                        resource_relation_func=(
                            json_home_relations.os_federation_resource_rel_func
                        ),
                    ),
                }
            ],
            rel='auth_domains',
            resource_kwargs={},
        ),
        ks_flask.construct_resource_map(
            resource=AuthSystemResource,
            url='/auth/system',
            resource_kwargs={},
            rel='auth_system',
        ),
        ks_flask.construct_resource_map(
            resource=AuthCatalogResource,
            url='/auth/catalog',
            resource_kwargs={},
            rel='auth_catalog',
        ),
        ks_flask.construct_resource_map(
            resource=AuthTokenOSPKIResource,
            url='/auth/tokens/OS-PKI/revoked',
            resource_kwargs={},
            rel='revocations',
            resource_relation_func=json_home_relations.os_pki_resource_rel_func,
        ),
        ks_flask.construct_resource_map(
            resource=AuthTokenResource,
            url='/auth/tokens',
            resource_kwargs={},
            rel='auth_tokens',
        ),
    ]


class AuthFederationAPI(ks_flask.APIBase):
    _name = 'auth/OS-FEDERATION'
    _import_name = __name__
    resources = []
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=AuthFederationSaml2Resource,
            url='/auth/OS-FEDERATION/saml2',
            resource_kwargs={},
            resource_relation_func=(
                json_home_relations.os_federation_resource_rel_func
            ),
            rel='saml2',
        ),
        ks_flask.construct_resource_map(
            resource=AuthFederationSaml2ECPResource,
            url='/auth/OS-FEDERATION/saml2/ecp',
            resource_kwargs={},
            resource_relation_func=(
                json_home_relations.os_federation_resource_rel_func
            ),
            rel='ecp',
        ),
        ks_flask.construct_resource_map(
            resource=AuthFederationWebSSOResource,
            url='/auth/OS-FEDERATION/websso/<string:protocol_id>',
            resource_kwargs={},
            rel='websso',
            resource_relation_func=(
                json_home_relations.os_federation_resource_rel_func
            ),
            path_vars={
                'protocol_id': (
                    json_home_relations.os_federation_parameter_rel_func(
                        parameter_name='protocol_id'
                    )
                )
            },
        ),
        ks_flask.construct_resource_map(
            resource=AuthFederationWebSSOIDPsResource,
            url=(
                '/auth/OS-FEDERATION/identity_providers/<string:idp_id>/'
                'protocols/<string:protocol_id>/websso'
            ),
            resource_kwargs={},
            rel='identity_providers_websso',
            resource_relation_func=(
                json_home_relations.os_federation_resource_rel_func
            ),
            path_vars={
                'idp_id': (
                    json_home_relations.os_federation_parameter_rel_func(
                        parameter_name='idp_id'
                    )
                ),
                'protocol_id': (
                    json_home_relations.os_federation_parameter_rel_func(
                        parameter_name='protocol_id'
                    )
                ),
            },
        ),
    ]


APIs = (AuthAPI, AuthFederationAPI)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_db import exception as db_exception

from keystone.assignment.role_backends import base
from keystone.assignment.role_backends import sql_model
from keystone.common import driver_hints
from keystone.common import resource_options
from keystone.common import sql
from keystone import exception


class Role(base.RoleDriverBase):
    @sql.handle_conflicts(conflict_type='role')
    def create_role(self, role_id, role):
        with sql.session_for_write() as session:
            ref = sql_model.RoleTable.from_dict(role)
            session.add(ref)
            # Set resource options passed on creation
            resource_options.resource_options_ref_to_mapper(
                ref, sql_model.RoleOption
            )
            return ref.to_dict()

    @driver_hints.truncated
    def list_roles(self, hints):
        # If there is a filter on domain_id and the value is None, then to
        # ensure that the sql filtering works correctly, we need to patch
        # the value to be NULL_DOMAIN_ID. This is safe to do here since we
        # know we are able to satisfy any filter of this type in the call to
        # filter_limit_query() below, which will remove the filter from the
        # hints (hence ensuring our substitution is not exposed to the caller).
        for f in hints.filters:
            if f['name'] == 'domain_id' and f['value'] is None:
                f['value'] = base.NULL_DOMAIN_ID

        with sql.session_for_read() as session:
            query = session.query(sql_model.RoleTable)
            refs = sql.filter_limit_query(sql_model.RoleTable, query, hints)
            return [ref.to_dict() for ref in refs]

    def list_roles_from_ids(self, ids):
        if not ids:
            return []
        else:
            with sql.session_for_read() as session:
                query = session.query(sql_model.RoleTable)
                query = query.filter(sql_model.RoleTable.id.in_(ids))
                role_refs = query.all()
                return [role_ref.to_dict() for role_ref in role_refs]

    def _get_role(self, session, role_id):
        ref = session.get(sql_model.RoleTable, role_id)
        if ref is None:
            raise exception.RoleNotFound(role_id=role_id)
        return ref

    def get_role(self, role_id):
        with sql.session_for_read() as session:
            return self._get_role(session, role_id).to_dict()

    @sql.handle_conflicts(conflict_type='role')
    def update_role(self, role_id, role):
        with sql.session_for_write() as session:
            ref = self._get_role(session, role_id)
            old_dict = ref.to_dict()
            for k in role:
                old_dict[k] = role[k]
            new_role = sql_model.RoleTable.from_dict(old_dict)
            for attr in sql_model.RoleTable.attributes:
                if attr != 'id':
                    setattr(ref, attr, getattr(new_role, attr))
            ref.extra = new_role.extra
            ref.description = new_role.description
            # Move the "_resource_options" attribute over to the real ref
            # so that resource_options.resource_options_ref_to_mapper can
            # handle the work.
            setattr(
                ref,
                '_resource_options',
                getattr(new_role, '_resource_options', {}),
            )
            # Move options into the propper attribute mapper construct
            resource_options.resource_options_ref_to_mapper(
                ref, sql_model.RoleOption
            )
            return ref.to_dict()

    def delete_role(self, role_id):
        with sql.session_for_write() as session:
            ref = self._get_role(session, role_id)
            session.delete(ref)

    def _get_implied_role(self, session, prior_role_id, implied_role_id):
        query = (
            session.query(sql_model.ImpliedRoleTable)
            .filter(sql_model.ImpliedRoleTable.prior_role_id == prior_role_id)
            .filter(
                sql_model.ImpliedRoleTable.implied_role_id == implied_role_id
            )
        )
        try:
            ref = query.one()
        except sql.NotFound:
            raise exception.ImpliedRoleNotFound(
                prior_role_id=prior_role_id, implied_role_id=implied_role_id
            )
        return ref

    @sql.handle_conflicts(conflict_type='implied_role')
    def create_implied_role(self, prior_role_id, implied_role_id):
        with sql.session_for_write() as session:
            inference = {
                'prior_role_id': prior_role_id,
                'implied_role_id': implied_role_id,
            }
            ref = sql_model.ImpliedRoleTable.from_dict(inference)
            try:
                session.add(ref)
            except db_exception.DBReferenceError:
                # We don't know which role threw this.
                # Query each to trigger the exception.
                self._get_role(session, prior_role_id)
                self._get_role(session, implied_role_id)
            return ref.to_dict()

    def delete_implied_role(self, prior_role_id, implied_role_id):
        with sql.session_for_write() as session:
            ref = self._get_implied_role(
                session, prior_role_id, implied_role_id
            )
            session.delete(ref)

    def list_implied_roles(self, prior_role_id):
        with sql.session_for_read() as session:
            query = session.query(sql_model.ImpliedRoleTable).filter(
                sql_model.ImpliedRoleTable.prior_role_id == prior_role_id
            )
            refs = query.all()
            return [ref.to_dict() for ref in refs]

    def list_role_inference_rules(self):
        with sql.session_for_read() as session:
            query = session.query(sql_model.ImpliedRoleTable)
            refs = query.all()
            return [ref.to_dict() for ref in refs]

    def get_implied_role(self, prior_role_id, implied_role_id):
        with sql.session_for_read() as session:
            ref = self._get_implied_role(
                session, prior_role_id, implied_role_id
            )
            return ref.to_dict()



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import datetime
import http.client
import typing as ty
import uuid

import oslo_context.context
from oslo_serialization import jsonutils
from testtools import matchers
import webtest

from keystone.common import authorization
from keystone.common import cache
from keystone.common import provider_api
from keystone.common.validation import validators
from keystone import exception
from keystone.resource.backends import base as resource_base
from keystone.server.flask.request_processing.middleware import auth_context
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import rest

PROVIDERS = provider_api.ProviderAPIs
DEFAULT_DOMAIN_ID = 'default'

TIME_FORMAT = unit.TIME_FORMAT


class RestfulTestCase(
    unit.SQLDriverOverrides, rest.RestfulTestCase, common_auth.AuthTestMixin
):
    def generate_token_schema(
        self, system_scoped=False, domain_scoped=False, project_scoped=False
    ):
        """Return a dictionary of token properties to validate against."""
        ROLES_SCHEMA = {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'id': {'type': 'string'},
                    'name': {'type': 'string'},
                    'description': {'type': 'string'},
                    'options': {'type': 'object'},
                },
                'required': ['id', 'name'],
                'additionalProperties': False,
            },
            'minItems': 1,
        }

        properties = {
            'audit_ids': {
                'type': 'array',
                'items': {'type': 'string'},
                'minItems': 1,
                'maxItems': 2,
            },
            'expires_at': {
                'type': 'string',
                'pattern': unit.TIME_FORMAT_REGEX,
            },
            'issued_at': {'type': 'string', 'pattern': unit.TIME_FORMAT_REGEX},
            'methods': {'type': 'array', 'items': {'type': 'string'}},
            'user': {
                'type': 'object',
                'required': ['id', 'name', 'domain', 'password_expires_at'],
                'properties': {
                    'id': {'type': 'string'},
                    'name': {'type': 'string'},
                    'domain': {
                        'type': 'object',
                        'properties': {
                            'id': {'type': 'string'},
                            'name': {'type': 'string'},
                        },
                        'required': ['id', 'name'],
                        'additonalProperties': False,
                    },
                    'password_expires_at': {
                        'type': ['string', 'null'],
                        'pattern': unit.TIME_FORMAT_REGEX,
                    },
                },
                'additionalProperties': False,
            },
        }

        if system_scoped:
            properties['catalog'] = {'type': 'array'}
            properties['system'] = {
                'type': 'object',
                'properties': {'all': {'type': 'boolean'}},
            }
            properties['roles'] = ROLES_SCHEMA
        elif domain_scoped:
            properties['catalog'] = {'type': 'array'}
            properties['roles'] = ROLES_SCHEMA
            properties['domain'] = {
                'type': 'object',
                'required': ['id', 'name'],
                'properties': {
                    'id': {'type': 'string'},
                    'name': {'type': 'string'},
                },
                'additionalProperties': False,
            }
        elif project_scoped:
            properties['is_admin_project'] = {'type': 'boolean'}
            properties['catalog'] = {'type': 'array'}
            # FIXME(lbragstad): Remove this in favor of the predefined
            # ROLES_SCHEMA dictionary once bug 1763510 is fixed.
            ROLES_SCHEMA['items']['properties']['domain_id'] = {
                'type': ['null', 'string']
            }
            properties['roles'] = ROLES_SCHEMA
            properties['is_domain'] = {'type': 'boolean'}
            properties['project'] = {
                'type': ['object'],
                'required': ['id', 'name', 'domain'],
                'properties': {
                    'id': {'type': 'string'},
                    'name': {'type': 'string'},
                    'domain': {
                        'type': ['object'],
                        'required': ['id', 'name'],
                        'properties': {
                            'id': {'type': 'string'},
                            'name': {'type': 'string'},
                        },
                        'additionalProperties': False,
                    },
                },
                'additionalProperties': False,
            }

        schema = {
            'type': 'object',
            'properties': properties,
            'required': [
                'audit_ids',
                'expires_at',
                'issued_at',
                'methods',
                'user',
            ],
            'optional': [],
            'additionalProperties': False,
        }

        if system_scoped:
            schema['required'].extend(['system', 'roles'])
            schema['optional'].append('catalog')
        elif domain_scoped:
            schema['required'].extend(['domain', 'roles'])
            schema['optional'].append('catalog')
        elif project_scoped:
            schema['required'].append('project')
            schema['optional'].append('catalog')
            schema['optional'].append('OS-TRUST:trust')
            schema['optional'].append('is_admin_project')
            schema['optional'].append('is_domain')

        return schema

    def config_files(self):
        config_files = super().config_files()
        config_files.append(unit.dirs.tests_conf('backend_sql.conf'))
        return config_files

    def setUp(self):
        """Setup for v3 Restful Test Cases."""
        super().setUp()

        self.empty_context = {'environment': {}}

    def load_backends(self):
        # ensure the cache region instance is setup
        cache.configure_cache()

        super().load_backends()

    def load_fixtures(self, fixtures):
        self.load_sample_data()

    def _populate_default_domain(self):
        try:
            PROVIDERS.resource_api.get_domain(DEFAULT_DOMAIN_ID)
        except exception.DomainNotFound:
            root_domain = unit.new_domain_ref(
                id=resource_base.NULL_DOMAIN_ID,
                name=resource_base.NULL_DOMAIN_ID,
            )
            PROVIDERS.resource_api.create_domain(
                resource_base.NULL_DOMAIN_ID, root_domain
            )
            domain = unit.new_domain_ref(
                description=('The default domain'),
                id=DEFAULT_DOMAIN_ID,
                name='Default',
            )
            PROVIDERS.resource_api.create_domain(DEFAULT_DOMAIN_ID, domain)

    def load_sample_data(self, create_region_and_endpoints=True):
        self._populate_default_domain()
        self.domain = unit.new_domain_ref()
        self.domain_id = self.domain['id']
        PROVIDERS.resource_api.create_domain(self.domain_id, self.domain)

        self.project = unit.new_project_ref(domain_id=self.domain_id)
        self.project_id = self.project['id']
        self.project = PROVIDERS.resource_api.create_project(
            self.project_id, self.project
        )

        self.user = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain_id
        )
        self.user_id = self.user['id']

        self.default_domain_project_id = uuid.uuid4().hex
        self.default_domain_project = unit.new_project_ref(
            domain_id=DEFAULT_DOMAIN_ID
        )
        self.default_domain_project['id'] = self.default_domain_project_id
        PROVIDERS.resource_api.create_project(
            self.default_domain_project_id, self.default_domain_project
        )

        self.default_domain_user = unit.create_user(
            PROVIDERS.identity_api, domain_id=DEFAULT_DOMAIN_ID
        )
        self.default_domain_user_id = self.default_domain_user['id']

        # create & grant policy.yaml's default role for admin_required
        self.role = unit.new_role_ref(name='admin')
        self.role_id = self.role['id']
        PROVIDERS.role_api.create_role(self.role_id, self.role)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.user_id, self.project_id, self.role_id
        )
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.default_domain_user_id,
            self.default_domain_project_id,
            self.role_id,
        )
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.default_domain_user_id, self.project_id, self.role_id
        )

        # Create "req_admin" user for simulating a real user instead of the
        # admin_token_auth middleware
        self.user_reqadmin = unit.create_user(
            PROVIDERS.identity_api, DEFAULT_DOMAIN_ID
        )
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.user_reqadmin['id'],
            self.default_domain_project_id,
            self.role_id,
        )

        if create_region_and_endpoints:
            self.region = unit.new_region_ref()
            self.region_id = self.region['id']
            PROVIDERS.catalog_api.create_region(self.region)

            self.service = unit.new_service_ref(type='identity')
            self.service_id = self.service['id']
            PROVIDERS.catalog_api.create_service(
                self.service_id, self.service.copy()
            )

            self.endpoint = unit.new_endpoint_ref(
                service_id=self.service_id,
                interface='public',
                region_id=self.region_id,
            )
            self.endpoint_id = self.endpoint['id']
            PROVIDERS.catalog_api.create_endpoint(
                self.endpoint_id, self.endpoint.copy()
            )
            # The server adds 'enabled' and defaults to True.
            self.endpoint['enabled'] = True

    def create_new_default_project_for_user(
        self, user_id, domain_id, enable_project=True
    ):
        ref = unit.new_project_ref(domain_id=domain_id, enabled=enable_project)
        r = self.post('/projects', body={'project': ref})
        project = self.assertValidProjectResponse(r, ref)
        # set the user's preferred project
        body = {'user': {'default_project_id': project['id']}}
        r = self.patch(f'/users/{user_id}', body=body)
        self.assertValidUserResponse(r)

        return project

    def get_admin_token(self):
        """Convenience method so that we can test authenticated requests."""
        r = self.admin_request(
            method='POST',
            path='/v3/auth/tokens',
            body={
                'auth': {
                    'identity': {
                        'methods': ['password'],
                        'password': {
                            'user': {
                                'name': self.user_reqadmin['name'],
                                'password': self.user_reqadmin['password'],
                                'domain': {
                                    'id': self.user_reqadmin['domain_id']
                                },
                            }
                        },
                    },
                    'scope': {
                        'project': {'id': self.default_domain_project_id}
                    },
                }
            },
        )
        return r.headers.get('X-Subject-Token')

    def get_unscoped_token(self):
        """Convenience method so that we can test authenticated requests."""
        r = self.admin_request(
            method='POST',
            path='/v3/auth/tokens',
            body={
                'auth': {
                    'identity': {
                        'methods': ['password'],
                        'password': {
                            'user': {
                                'name': self.user['name'],
                                'password': self.user['password'],
                                'domain': {'id': self.user['domain_id']},
                            }
                        },
                    }
                }
            },
        )
        return r.headers.get('X-Subject-Token')

    def get_scoped_token(self):
        """Convenience method so that we can test authenticated requests."""
        r = self.admin_request(
            method='POST',
            path='/v3/auth/tokens',
            body={
                'auth': {
                    'identity': {
                        'methods': ['password'],
                        'password': {
                            'user': {
                                'name': self.user['name'],
                                'password': self.user['password'],
                                'domain': {'id': self.user['domain_id']},
                            }
                        },
                    },
                    'scope': {'project': {'id': self.project['id']}},
                }
            },
        )
        return r.headers.get('X-Subject-Token')

    def get_system_scoped_token(self):
        """Convenience method for requesting system scoped tokens."""
        r = self.admin_request(
            method='POST',
            path='/v3/auth/tokens',
            body={
                'auth': {
                    'identity': {
                        'methods': ['password'],
                        'password': {
                            'user': {
                                'name': self.user['name'],
                                'password': self.user['password'],
                                'domain': {'id': self.user['domain_id']},
                            }
                        },
                    },
                    'scope': {'system': {'all': True}},
                }
            },
        )
        return r.headers.get('X-Subject-Token')

    def get_domain_scoped_token(self):
        """Convenience method for requesting domain scoped token."""
        r = self.admin_request(
            method='POST',
            path='/v3/auth/tokens',
            body={
                'auth': {
                    'identity': {
                        'methods': ['password'],
                        'password': {
                            'user': {
                                'name': self.user['name'],
                                'password': self.user['password'],
                                'domain': {'id': self.user['domain_id']},
                            }
                        },
                    },
                    'scope': {'domain': {'id': self.domain['id']}},
                }
            },
        )
        return r.headers.get('X-Subject-Token')

    def get_application_credentials_token(self, app_cred_id, app_cred_secret):
        """Convenience method for requesting application credentials token."""
        r = self.admin_request(
            method='POST',
            path='/v3/auth/tokens',
            body={
                'auth': {
                    'identity': {
                        'methods': ['application_credential'],
                        'application_credential': {
                            'id': app_cred_id,
                            'secret': app_cred_secret,
                        },
                    }
                }
            },
        )
        return r.headers.get('X-Subject-Token')

    def get_requested_token(self, auth):
        """Request the specific token we want."""
        r = self.v3_create_token(auth)
        return r.headers.get('X-Subject-Token')

    def v3_create_token(self, auth, expected_status=http.client.CREATED):
        return self.admin_request(
            method='POST',
            path='/v3/auth/tokens',
            body=auth,
            expected_status=expected_status,
        )

    def v3_noauth_request(self, path, **kwargs):
        # request does not require auth token header
        path = '/v3' + path
        return self.admin_request(path=path, **kwargs)

    def v3_request(self, path, **kwargs):
        # check to see if caller requires token for the API call.
        if kwargs.pop('noauth', None):
            return self.v3_noauth_request(path, **kwargs)

        # Check if the caller has passed in auth details for
        # use in requesting the token
        auth_arg = kwargs.pop('auth', None)
        if auth_arg:
            token = self.get_requested_token(auth_arg)
        else:
            token = kwargs.pop('token', None)
            if not token:
                token = self.get_scoped_token()
        path = '/v3' + path

        return self.admin_request(path=path, token=token, **kwargs)

    def get(self, path, expected_status=http.client.OK, **kwargs):
        return self.v3_request(
            path, method='GET', expected_status=expected_status, **kwargs
        )

    def head(self, path, expected_status=http.client.NO_CONTENT, **kwargs):
        r = self.v3_request(
            path, method='HEAD', expected_status=expected_status, **kwargs
        )
        self.assertEqual(b'', r.body)
        return r

    def post(self, path, expected_status=http.client.CREATED, **kwargs):
        return self.v3_request(
            path, method='POST', expected_status=expected_status, **kwargs
        )

    def put(self, path, expected_status=http.client.NO_CONTENT, **kwargs):
        return self.v3_request(
            path, method='PUT', expected_status=expected_status, **kwargs
        )

    def patch(self, path, expected_status=http.client.OK, **kwargs):
        return self.v3_request(
            path, method='PATCH', expected_status=expected_status, **kwargs
        )

    def delete(self, path, expected_status=http.client.NO_CONTENT, **kwargs):
        return self.v3_request(
            path, method='DELETE', expected_status=expected_status, **kwargs
        )

    def assertValidErrorResponse(self, r):
        resp = r.result
        self.assertIsNotNone(resp.get('error'))
        self.assertIsNotNone(resp['error'].get('code'))
        self.assertIsNotNone(resp['error'].get('title'))
        self.assertIsNotNone(resp['error'].get('message'))
        self.assertEqual(int(resp['error']['code']), r.status_code)

    def assertValidListLinks(self, links, resource_url=None):
        self.assertIsNotNone(links)
        self.assertIsNotNone(links.get('self'))
        self.assertThat(links['self'], matchers.StartsWith('http://localhost'))

        if resource_url:
            self.assertThat(links['self'], matchers.EndsWith(resource_url))

        self.assertIn('next', links)
        if links['next'] is not None:
            self.assertThat(
                links['next'], matchers.StartsWith('http://localhost')
            )

        self.assertIn('previous', links)
        if links['previous'] is not None:
            self.assertThat(
                links['previous'], matchers.StartsWith('http://localhost')
            )

    def assertValidListResponse(
        self,
        resp,
        key,
        entity_validator,
        ref=None,
        expected_length=None,
        keys_to_check=None,
        resource_url=None,
    ):
        """Make assertions common to all API list responses.

        If a reference is provided, it's ID will be searched for in the
        response, and asserted to be equal.

        """
        entities = resp.result.get(key)
        self.assertIsNotNone(entities)

        if expected_length is not None:
            self.assertEqual(expected_length, len(entities))
        elif ref is not None:
            # we're at least expecting the ref
            self.assertNotEmpty(entities)

        # collections should have relational links
        self.assertValidListLinks(
            resp.result.get('links'), resource_url=resource_url
        )

        for entity in entities:
            self.assertIsNotNone(entity)
            self.assertValidEntity(entity, keys_to_check=keys_to_check)
            entity_validator(entity)
        if ref:
            entity = [x for x in entities if x['id'] == ref['id']][0]
            self.assertValidEntity(
                entity, ref=ref, keys_to_check=keys_to_check
            )
            entity_validator(entity, ref)
        return entities

    def assertValidResponse(
        self, resp, key, entity_validator, *args, **kwargs
    ):
        """Make assertions common to all API responses."""
        entity = resp.result.get(key)
        self.assertIsNotNone(entity)
        keys = kwargs.pop('keys_to_check', None)
        self.assertValidEntity(entity, keys_to_check=keys, *args, **kwargs)
        entity_validator(entity, *args, **kwargs)
        return entity

    def assertValidEntity(self, entity, ref=None, keys_to_check=None):
        """Make assertions common to all API entities.

        If a reference is provided, the entity will also be compared against
        the reference.
        """
        if keys_to_check is not None:
            keys = keys_to_check
        else:
            keys = ['name', 'description', 'enabled']

        for k in ['id'] + keys:
            msg = f'{k} unexpectedly None in {entity}'
            self.assertIsNotNone(entity.get(k), msg)

        self.assertIsNotNone(entity.get('links'))
        self.assertIsNotNone(entity['links'].get('self'))
        self.assertThat(
            entity['links']['self'], matchers.StartsWith('http://localhost')
        )
        self.assertIn(entity['id'], entity['links']['self'])

        if ref:
            for k in keys:
                msg = f'{k} not equal: {ref[k]} != {entity[k]}'
                self.assertEqual(ref[k], entity[k])

        return entity

    # auth validation

    def assertValidISO8601ExtendedFormatDatetime(self, dt):
        try:
            return datetime.datetime.strptime(dt, TIME_FORMAT)
        except Exception:
            msg = f'{dt} is not a valid ISO 8601 extended format date time.'
            raise AssertionError(msg)

    def assertValidTokenResponse(self, r, user=None, forbid_token_id=False):
        if forbid_token_id:
            self.assertNotIn('X-Subject-Token', r.headers)
        else:
            self.assertTrue(r.headers.get('X-Subject-Token'))
        token = r.result['token']

        self.assertIsNotNone(token.get('expires_at'))
        expires_at = self.assertValidISO8601ExtendedFormatDatetime(
            token['expires_at']
        )
        self.assertIsNotNone(token.get('issued_at'))
        issued_at = self.assertValidISO8601ExtendedFormatDatetime(
            token['issued_at']
        )
        self.assertLess(issued_at, expires_at)

        self.assertIn('user', token)
        self.assertIn('id', token['user'])
        self.assertIn('name', token['user'])
        self.assertIn('domain', token['user'])
        self.assertIn('id', token['user']['domain'])

        if user is not None:
            self.assertEqual(user['id'], token['user']['id'])
            self.assertEqual(user['name'], token['user']['name'])
            self.assertEqual(user['domain_id'], token['user']['domain']['id'])

        return token

    def assertValidUnscopedTokenResponse(self, r, *args, **kwargs):
        token = self.assertValidTokenResponse(r, *args, **kwargs)
        validator_object = validators.SchemaValidator(
            self.generate_token_schema()
        )
        validator_object.validate(token)

        return token

    def assertValidScopedTokenResponse(self, r, *args, **kwargs):
        require_catalog = kwargs.pop('require_catalog', True)
        endpoint_filter = kwargs.pop('endpoint_filter', False)
        ep_filter_assoc = kwargs.pop('ep_filter_assoc', 0)
        is_admin_project = kwargs.pop('is_admin_project', None)
        token = self.assertValidTokenResponse(r, *args, **kwargs)

        if require_catalog:
            endpoint_num = 0
            self.assertIn('catalog', token)

            if isinstance(token['catalog'], list):
                # only test JSON
                for service in token['catalog']:
                    for endpoint in service['endpoints']:
                        self.assertNotIn('enabled', endpoint)
                        self.assertNotIn('legacy_endpoint_id', endpoint)
                        self.assertNotIn('service_id', endpoint)
                        endpoint_num += 1

            # sub test for the OS-EP-FILTER extension enabled
            if endpoint_filter:
                self.assertEqual(ep_filter_assoc, endpoint_num)
        else:
            self.assertNotIn('catalog', token)

        self.assertIn('roles', token)
        self.assertTrue(token['roles'])
        for role in token['roles']:
            self.assertIn('id', role)
            self.assertIn('name', role)

        # NOTE(samueldmq): We want to explicitly test for boolean or None
        self.assertIs(is_admin_project, token.get('is_admin_project'))

        return token

    def assertValidProjectScopedTokenResponse(self, r, *args, **kwargs):
        token = self.assertValidScopedTokenResponse(r, *args, **kwargs)

        project_scoped_token_schema = self.generate_token_schema(
            project_scoped=True
        )

        if token.get('OS-TRUST:trust'):
            trust_properties = {
                'OS-TRUST:trust': {
                    'type': ['object'],
                    'required': [
                        'id',
                        'impersonation',
                        'trustor_user',
                        'trustee_user',
                    ],
                    'properties': {
                        'id': {'type': 'string'},
                        'impersonation': {'type': 'boolean'},
                        'trustor_user': {
                            'type': 'object',
                            'required': ['id'],
                            'properties': {'id': {'type': 'string'}},
                            'additionalProperties': False,
                        },
                        'trustee_user': {
                            'type': 'object',
                            'required': ['id'],
                            'properties': {'id': {'type': 'string'}},
                            'additionalProperties': False,
                        },
                    },
                    'additionalProperties': False,
                }
            }
            project_scoped_token_schema['properties'].update(trust_properties)

        validator_object = validators.SchemaValidator(
            project_scoped_token_schema
        )
        validator_object.validate(token)

        self.assertEqual(self.role_id, token['roles'][0]['id'])

        return token

    def assertValidDomainScopedTokenResponse(self, r, *args, **kwargs):
        token = self.assertValidScopedTokenResponse(r, *args, **kwargs)

        validator_object = validators.SchemaValidator(
            self.generate_token_schema(domain_scoped=True)
        )
        validator_object.validate(token)

        return token

    def assertValidSystemScopedTokenResponse(self, r, *args, **kwargs):
        token = self.assertValidTokenResponse(r)
        self.assertTrue(token['system']['all'])

        system_scoped_token_schema = self.generate_token_schema(
            system_scoped=True
        )
        validator_object = validators.SchemaValidator(
            system_scoped_token_schema
        )
        validator_object.validate(token)

        return token

    # catalog validation

    def assertValidCatalogResponse(self, resp, *args, **kwargs):
        self.assertEqual({'catalog', 'links'}, set(resp.json.keys()))
        self.assertValidCatalog(resp.json['catalog'])
        self.assertIn('links', resp.json)
        self.assertIsInstance(resp.json['links'], dict)
        self.assertEqual(['self'], list(resp.json['links'].keys()))
        self.assertEqual(
            'http://localhost/v3/auth/catalog', resp.json['links']['self']
        )

    def assertValidCatalog(self, entity):
        self.assertIsInstance(entity, list)
        self.assertGreater(len(entity), 0)
        for service in entity:
            self.assertIsNotNone(service.get('id'))
            self.assertIsNotNone(service.get('name'))
            self.assertIsNotNone(service.get('type'))
            self.assertNotIn('enabled', service)
            self.assertGreater(len(service['endpoints']), 0)
            for endpoint in service['endpoints']:
                self.assertIsNotNone(endpoint.get('id'))
                self.assertIsNotNone(endpoint.get('interface'))
                self.assertIsNotNone(endpoint.get('url'))
                self.assertNotIn('enabled', endpoint)
                self.assertNotIn('legacy_endpoint_id', endpoint)
                self.assertNotIn('service_id', endpoint)

    # region validation

    def assertValidRegionListResponse(self, resp, *args, **kwargs):
        # NOTE(jaypipes): I have to pass in a blank keys_to_check parameter
        #                 below otherwise the base assertValidEntity method
        #                 tries to find a "name" and an "enabled" key in the
        #                 returned ref dicts. The issue is, I don't understand
        #                 how the service and endpoint entity assertions below
        #                 actually work (they don't raise assertions), since
        #                 AFAICT, the service and endpoint tables don't have
        #                 a "name" column either... :(
        return self.assertValidListResponse(
            resp,
            'regions',
            self.assertValidRegion,
            keys_to_check=[],
            *args,
            **kwargs,
        )

    def assertValidRegionResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp,
            'region',
            self.assertValidRegion,
            keys_to_check=[],
            *args,
            **kwargs,
        )

    def assertValidRegion(self, entity, ref=None):
        self.assertIsNotNone(entity.get('description'))
        if ref:
            self.assertEqual(ref['description'], entity['description'])
        return entity

    # service validation

    def assertValidServiceListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp, 'services', self.assertValidService, *args, **kwargs
        )

    def assertValidServiceResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp, 'service', self.assertValidService, *args, **kwargs
        )

    def assertValidService(self, entity, ref=None):
        self.assertIsNotNone(entity.get('type'))
        self.assertIsInstance(entity.get('enabled'), bool)
        if ref:
            self.assertEqual(ref['type'], entity['type'])
        return entity

    # endpoint validation

    def assertValidEndpointListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp, 'endpoints', self.assertValidEndpoint, *args, **kwargs
        )

    def assertValidEndpointResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp, 'endpoint', self.assertValidEndpoint, *args, **kwargs
        )

    def assertValidEndpoint(self, entity, ref=None):
        self.assertIsNotNone(entity.get('interface'))
        self.assertIsNotNone(entity.get('service_id'))
        self.assertIsInstance(entity['enabled'], bool)

        # this is intended to be an unexposed implementation detail
        self.assertNotIn('legacy_endpoint_id', entity)

        if ref:
            self.assertEqual(ref['interface'], entity['interface'])
            self.assertEqual(ref['service_id'], entity['service_id'])
            if ref.get('region') is not None:
                self.assertEqual(ref['region_id'], entity.get('region_id'))

        return entity

    # domain validation

    def assertValidDomainListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp, 'domains', self.assertValidDomain, *args, **kwargs
        )

    def assertValidDomainResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp, 'domain', self.assertValidDomain, *args, **kwargs
        )

    def assertValidDomain(self, entity, ref=None):
        if ref:
            pass
        return entity

    # project validation

    def assertValidProjectListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp, 'projects', self.assertValidProject, *args, **kwargs
        )

    def assertValidProjectResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp, 'project', self.assertValidProject, *args, **kwargs
        )

    def assertValidProject(self, entity, ref=None):
        if ref:
            self.assertEqual(ref['domain_id'], entity['domain_id'])
        return entity

    # user validation

    def assertValidUserListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp,
            'users',
            self.assertValidUser,
            keys_to_check=['name', 'enabled'],
            *args,
            **kwargs,
        )

    def assertValidUserResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp,
            'user',
            self.assertValidUser,
            keys_to_check=['name', 'enabled'],
            *args,
            **kwargs,
        )

    def assertValidUser(self, entity, ref=None):
        self.assertIsNotNone(entity.get('domain_id'))
        self.assertIsNotNone(entity.get('email'))
        self.assertNotIn('password', entity)
        self.assertNotIn('projectId', entity)
        self.assertIn('password_expires_at', entity)
        if ref:
            self.assertEqual(ref['domain_id'], entity['domain_id'])
            self.assertEqual(ref['email'], entity['email'])
            if 'default_project_id' in ref:
                self.assertIsNotNone(ref['default_project_id'])
                self.assertEqual(
                    ref['default_project_id'], entity['default_project_id']
                )
        return entity

    # group validation

    def assertValidGroupListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp,
            'groups',
            self.assertValidGroup,
            keys_to_check=['name', 'description', 'domain_id'],
            *args,
            **kwargs,
        )

    def assertValidGroupResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp,
            'group',
            self.assertValidGroup,
            keys_to_check=['name', 'description', 'domain_id'],
            *args,
            **kwargs,
        )

    def assertValidGroup(self, entity, ref=None):
        self.assertIsNotNone(entity.get('name'))
        if ref:
            self.assertEqual(ref['name'], entity['name'])
        return entity

    # credential validation

    def assertValidCredentialListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp,
            'credentials',
            self.assertValidCredential,
            keys_to_check=['blob', 'user_id', 'type'],
            *args,
            **kwargs,
        )

    def assertValidCredentialResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp,
            'credential',
            self.assertValidCredential,
            keys_to_check=['blob', 'user_id', 'type'],
            *args,
            **kwargs,
        )

    def assertValidCredential(self, entity, ref=None):
        self.assertIsNotNone(entity.get('user_id'))
        self.assertIsNotNone(entity.get('blob'))
        self.assertIsNotNone(entity.get('type'))
        self.assertNotIn('key_hash', entity)
        self.assertNotIn('encrypted_blob', entity)
        if ref:
            self.assertEqual(ref['user_id'], entity['user_id'])
            self.assertEqual(ref['blob'], entity['blob'])
            self.assertEqual(ref['type'], entity['type'])
            self.assertEqual(ref.get('project_id'), entity.get('project_id'))
        return entity

    # role validation

    def assertValidRoleListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp,
            'roles',
            self.assertValidRole,
            keys_to_check=['name'],
            *args,
            **kwargs,
        )

    def assertRoleInListResponse(self, resp, ref, expected=1):
        found_count = 0
        for entity in resp.result.get('roles'):
            try:
                self.assertValidRole(entity, ref=ref)
            except Exception:
                # It doesn't match, so let's go onto the next one
                pass
            else:
                found_count += 1
        self.assertEqual(expected, found_count)

    def assertRoleNotInListResponse(self, resp, ref):
        self.assertRoleInListResponse(resp, ref=ref, expected=0)

    def assertValidRoleResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp,
            'role',
            self.assertValidRole,
            keys_to_check=['name'],
            *args,
            **kwargs,
        )

    def assertValidRole(self, entity, ref=None):
        self.assertIsNotNone(entity.get('name'))
        if ref:
            self.assertEqual(ref['name'], entity['name'])
            self.assertEqual(ref['domain_id'], entity['domain_id'])
        return entity

    # role assignment validation

    def assertValidRoleAssignmentListResponse(
        self, resp, expected_length=None, resource_url=None
    ):
        entities = resp.result.get('role_assignments')

        if expected_length or expected_length == 0:
            self.assertEqual(expected_length, len(entities))

        # Collections should have relational links
        self.assertValidListLinks(
            resp.result.get('links'), resource_url=resource_url
        )

        for entity in entities:
            self.assertIsNotNone(entity)
            self.assertValidRoleAssignment(entity)
        return entities

    def assertValidRoleAssignment(self, entity, ref=None):
        # A role should be present
        self.assertIsNotNone(entity.get('role'))
        self.assertIsNotNone(entity['role'].get('id'))

        # Only one of user or group should be present
        if entity.get('user'):
            self.assertNotIn('group', entity)
            self.assertIsNotNone(entity['user'].get('id'))
        else:
            self.assertIsNotNone(entity.get('group'))
            self.assertIsNotNone(entity['group'].get('id'))

        # A scope should be present and have only one of domain or project
        self.assertIsNotNone(entity.get('scope'))

        if entity['scope'].get('project'):
            self.assertNotIn('domain', entity['scope'])
            self.assertIsNotNone(entity['scope']['project'].get('id'))
        elif entity['scope'].get('domain'):
            self.assertIsNotNone(entity['scope'].get('domain'))
            self.assertIsNotNone(entity['scope']['domain'].get('id'))
        else:
            self.assertIsNotNone(entity['scope'].get('system'))
            self.assertTrue(entity['scope']['system']['all'])

        # An assignment link should be present
        self.assertIsNotNone(entity.get('links'))
        self.assertIsNotNone(entity['links'].get('assignment'))

        if ref:
            links = ref.pop('links')
            try:
                self.assertLessEqual(ref.items(), entity.items())
                self.assertIn(
                    links['assignment'], entity['links']['assignment']
                )
            finally:
                if links:
                    ref['links'] = links

    def assertRoleAssignmentInListResponse(self, resp, ref, expected=1):
        found_count = 0
        for entity in resp.result.get('role_assignments'):
            try:
                self.assertValidRoleAssignment(entity, ref=ref)
            except Exception:
                # It doesn't match, so let's go onto the next one
                pass
            else:
                found_count += 1
        self.assertEqual(expected, found_count)

    def assertRoleAssignmentNotInListResponse(self, resp, ref):
        self.assertRoleAssignmentInListResponse(resp, ref=ref, expected=0)

    # policy validation

    def assertValidPolicyListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp, 'policies', self.assertValidPolicy, *args, **kwargs
        )

    def assertValidPolicyResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp, 'policy', self.assertValidPolicy, *args, **kwargs
        )

    def assertValidPolicy(self, entity, ref=None):
        self.assertIsNotNone(entity.get('blob'))
        self.assertIsNotNone(entity.get('type'))
        if ref:
            self.assertEqual(ref['blob'], entity['blob'])
            self.assertEqual(ref['type'], entity['type'])
        return entity

    # trust validation

    def assertValidTrustListResponse(self, resp, *args, **kwargs):
        return self.assertValidListResponse(
            resp,
            'trusts',
            self.assertValidTrustSummary,
            keys_to_check=[
                'trustor_user_id',
                'trustee_user_id',
                'impersonation',
            ],
            *args,
            **kwargs,
        )

    def assertValidTrustResponse(self, resp, *args, **kwargs):
        return self.assertValidResponse(
            resp,
            'trust',
            self.assertValidTrust,
            keys_to_check=[
                'trustor_user_id',
                'trustee_user_id',
                'impersonation',
            ],
            *args,
            **kwargs,
        )

    def assertValidTrustSummary(self, entity, ref=None):
        return self.assertValidTrust(entity, ref, summary=True)

    def assertValidTrust(self, entity, ref=None, summary=False):
        self.assertIsNotNone(entity.get('trustor_user_id'))
        self.assertIsNotNone(entity.get('trustee_user_id'))
        self.assertIsNotNone(entity.get('impersonation'))

        self.assertIn('expires_at', entity)
        if entity['expires_at'] is not None:
            self.assertValidISO8601ExtendedFormatDatetime(entity['expires_at'])

        if summary:
            # Trust list contains no roles, but getting a specific
            # trust by ID provides the detailed response containing roles
            self.assertNotIn('roles', entity)
            self.assertIn('project_id', entity)
        else:
            for role in entity['roles']:
                self.assertIsNotNone(role)
                self.assertValidEntity(role, keys_to_check=['name'])
                self.assertValidRole(role)

            self.assertValidListLinks(entity.get('roles_links'))
            # Mirror the test tempest does to ensure the self-link is correct
            self.assertIn('v3/OS-TRUST/trusts', entity.get('links')['self'])

            # always disallow role xor project_id (neither or both is allowed)
            has_roles = bool(entity.get('roles'))
            has_project = bool(entity.get('project_id'))
            self.assertFalse(has_roles ^ has_project)

        if ref:
            self.assertEqual(ref['trustor_user_id'], entity['trustor_user_id'])
            self.assertEqual(ref['trustee_user_id'], entity['trustee_user_id'])
            self.assertEqual(ref['project_id'], entity['project_id'])
            if entity.get('expires_at') or ref.get('expires_at'):
                entity_exp = self.assertValidISO8601ExtendedFormatDatetime(
                    entity['expires_at']
                )
                ref_exp = self.assertValidISO8601ExtendedFormatDatetime(
                    ref['expires_at']
                )
                self.assertCloseEnoughForGovernmentWork(entity_exp, ref_exp)
            else:
                self.assertEqual(
                    ref.get('expires_at'), entity.get('expires_at')
                )

        return entity

    # Service providers (federation)

    def assertValidServiceProvider(self, entity, ref=None, *args, **kwargs):
        attributes = frozenset(
            [
                'auth_url',
                'id',
                'enabled',
                'description',
                'links',
                'relay_state_prefix',
                'sp_url',
            ]
        )
        for attribute in attributes:
            self.assertIsNotNone(entity.get(attribute))

    def build_external_auth_environ(self, remote_user, remote_domain=None):
        environment = {'REMOTE_USER': remote_user, 'AUTH_TYPE': 'Negotiate'}
        if remote_domain:
            environment['REMOTE_DOMAIN'] = remote_domain
        return environment


class OAuth2RestfulTestCase(RestfulTestCase):
    def assertValidErrorResponse(self, response):
        resp = response.result
        self.assertIsNotNone(resp.get('error'))
        self.assertIsNotNone(resp.get('error_description'))


class VersionTestCase(RestfulTestCase):
    def test_get_version(self):
        pass


# NOTE(gyee): test AuthContextMiddleware here instead of test_middleware.py
# because we need the token
class AuthContextMiddlewareTestCase(RestfulTestCase):
    def load_fixtures(self, fixtures):
        self.load_sample_data()

        app_cred_api = PROVIDERS.application_credential_api
        access_rules = [
            {
                'id': uuid.uuid4().hex,
                'service': self.service['type'],
                'method': 'GET',
                'path': '/v3/users/*',
            }
        ]
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': 'appcredtest',
            'secret': uuid.uuid4().hex,
            'user_id': self.user['id'],
            'project_id': self.project['id'],
            'description': 'Test Application Credential',
            'roles': [{'id': self.role_id}],
            'access_rules': access_rules,
        }
        app_cred_ref = app_cred_api.create_application_credential(app_cred)
        self.app_cred_r_id = app_cred_ref['id']
        self.app_cred_r_secret = app_cred_ref['secret']

    def _middleware_request(self, token, extra_environ=None):
        def application(environ, start_response):
            body = b'body'
            headers = [
                ('Content-Type', 'text/html; charset=utf8'),
                ('Content-Length', str(len(body))),
            ]
            start_response('200 OK', headers)
            return [body]

        app = webtest.TestApp(
            auth_context.AuthContextMiddleware(application),
            extra_environ=extra_environ,
        )
        resp = app.get('/', headers={authorization.AUTH_TOKEN_HEADER: token})
        self.assertEqual(b'body', resp.body)  # just to make sure it worked
        return resp.request

    def test_auth_context_build_by_middleware(self):
        # test to make sure AuthContextMiddleware successful build the auth
        # context from the incoming auth token
        admin_token = self.get_scoped_token()
        req = self._middleware_request(admin_token)
        self.assertEqual(
            self.user['id'],
            req.environ.get(authorization.AUTH_CONTEXT_ENV)['user_id'],
        )

    def test_auth_context_override(self):
        overridden_context = 'OVERRIDDEN_CONTEXT'
        # this token should not be used
        token = uuid.uuid4().hex

        extra_environ = {authorization.AUTH_CONTEXT_ENV: overridden_context}
        req = self._middleware_request(token, extra_environ=extra_environ)
        # make sure overridden context take precedence
        self.assertEqual(
            overridden_context, req.environ.get(authorization.AUTH_CONTEXT_ENV)
        )

    def test_unscoped_token_auth_context(self):
        unscoped_token = self.get_unscoped_token()
        req = self._middleware_request(unscoped_token)
        # This check originally looked that the value was unset
        # but that was an artifact of the custom context keystone
        # used to create.  Oslo-context will always provide the
        # same set of keys, but the values will be None in an
        # unscoped token
        for key in ['project_id', 'domain_id', 'domain_name']:
            self.assertIsNone(
                req.environ.get(authorization.AUTH_CONTEXT_ENV)[key]
            )

    def test_project_scoped_token_auth_context(self):
        project_scoped_token = self.get_scoped_token()
        req = self._middleware_request(project_scoped_token)
        self.assertEqual(
            self.project['id'],
            req.environ.get(authorization.AUTH_CONTEXT_ENV)['project_id'],
        )

    def test_domain_scoped_token_auth_context(self):
        # grant the domain role to user
        path = '/domains/{}/users/{}/roles/{}'.format(
            self.domain['id'], self.user['id'], self.role['id']
        )
        self.put(path=path)

        domain_scoped_token = self.get_domain_scoped_token()
        req = self._middleware_request(domain_scoped_token)
        self.assertEqual(
            self.domain['id'],
            req.environ.get(authorization.AUTH_CONTEXT_ENV)['domain_id'],
        )
        self.assertEqual(
            self.domain['name'],
            req.environ.get(authorization.AUTH_CONTEXT_ENV)['domain_name'],
        )

    def test_oslo_context(self):
        # After AuthContextMiddleware runs, an
        # oslo_context.context.RequestContext was created so that its fields
        # can be logged. This test validates that the RequestContext was
        # created and the fields are set as expected.

        # Use a scoped token so more fields can be set.
        token = self.get_scoped_token()

        # oslo_middleware RequestId middleware sets openstack.request_id.
        request_id = uuid.uuid4().hex
        environ = {'openstack.request_id': request_id}
        self._middleware_request(token, extra_environ=environ)

        req_context = oslo_context.context.get_current()
        self.assertEqual(request_id, req_context.request_id)
        self.assertEqual(token, req_context.auth_token)
        self.assertEqual(self.user['id'], req_context.user_id)
        self.assertEqual(self.project['id'], req_context.project_id)
        self.assertIsNone(req_context.domain_id)
        self.assertEqual(self.user['domain_id'], req_context.user_domain_id)
        self.assertEqual(
            self.project['domain_id'], req_context.project_domain_id
        )
        self.assertFalse(req_context.is_admin)

    def test_auth_context_app_cred_with_rule(self):
        #
        # This is an open-coded _middleware_request(), which allows us to
        # supply paths and verify failure. We can refactor later if needed.
        #
        def application(environ, start_response):
            body = b'body'
            headers = [
                ('Content-Type', 'text/html; charset=utf8'),
                ('Content-Length', str(len(body))),
            ]
            start_response('200 OK', headers)
            return [body]

        token = self.get_application_credentials_token(
            self.app_cred_r_id, self.app_cred_r_secret
        )

        # Test to failure
        app = webtest.TestApp(auth_context.AuthContextMiddleware(application))
        resp = app.get(
            '/v3/projects/e3a0883d15ff409e98e59d460f583a68',
            headers={authorization.AUTH_TOKEN_HEADER: token},
            status=401,
        )
        self.assertEqual('401 Unauthorized', resp.status)

        # Test to success
        app = webtest.TestApp(auth_context.AuthContextMiddleware(application))
        resp = app.get(
            '/v3/users/3879328537914be2b394ddf57a4fc73a',
            headers={authorization.AUTH_TOKEN_HEADER: token},
        )
        self.assertEqual('200 OK', resp.status)
        self.assertEqual(b'body', resp.body)  # just to make sure it worked


class JsonHomeTestMixin:
    """JSON Home test.

    Mixin this class to provide a test for the JSON-Home response for an
    extension.

    The base class must set JSON_HOME_DATA to a dict of relationship URLs
    (rels) to the JSON-Home data for the relationship. The rels and associated
    data must be in the response.

    """

    def test_get_json_home(self):
        resp = self.get(
            '/', convert=False, headers={'Accept': 'application/json-home'}
        )
        self.assertThat(
            resp.headers['Content-Type'],
            matchers.Equals('application/json-home'),
        )
        resp_data = jsonutils.loads(resp.body)

        # Check that the example relationships are present.
        for rel in self.JSON_HOME_DATA:
            self.assertThat(
                resp_data['resources'][rel],
                matchers.Equals(self.JSON_HOME_DATA[rel]),
            )


class AssignmentTestMixin:
    """To hold assignment helper functions."""

    def build_role_assignment_query_url(self, effective=False, **filters):
        """Build and return a role assignment query url with provided params.

        Available filters are: domain_id, project_id, user_id, group_id,
        role_id and inherited_to_projects.
        """
        query_params = '?effective' if effective else ''

        for k, v in filters.items():
            query_params += '?' if not query_params else '&'

            if k == 'inherited_to_projects':
                query_params += 'scope.OS-INHERIT:inherited_to=projects'
            else:
                if k in ['domain_id', 'project_id']:
                    query_params += 'scope.'
                elif k not in ['user_id', 'group_id', 'role_id']:
                    raise ValueError(
                        f'Invalid key \'{k}\' in provided filters.'
                    )

                query_params += '{}={}'.format(k.replace('_', '.'), v)

        return f'/role_assignments{query_params}'

    def build_role_assignment_link(self, **attribs):
        """Build and return a role assignment link with provided attributes.

        Provided attributes are expected to contain: domain_id or project_id,
        user_id or group_id, role_id and, optionally, inherited_to_projects.
        """
        if attribs.get('domain_id'):
            link = '/domains/' + attribs['domain_id']
        elif attribs.get('system'):
            link = '/system'
        else:
            link = '/projects/' + attribs['project_id']

        if attribs.get('user_id'):
            link += '/users/' + attribs['user_id']
        else:
            link += '/groups/' + attribs['group_id']

        link += '/roles/' + attribs['role_id']

        if attribs.get('inherited_to_projects'):
            return f'/OS-INHERIT{link}/inherited_to_projects'

        return link

    def build_role_assignment_entity(
        self, link=None, prior_role_link=None, **attribs
    ):
        """Build and return a role assignment entity with provided attributes.

        Provided attributes are expected to contain: domain_id or project_id,
        user_id or group_id, role_id and, optionally, inherited_to_projects.
        """
        entity = {
            'links': {
                'assignment': (
                    link or self.build_role_assignment_link(**attribs)
                )
            }
        }

        if attribs.get('domain_id'):
            entity['scope'] = {'domain': {'id': attribs['domain_id']}}
        elif attribs.get('system'):
            entity['scope'] = {'system': {'all': True}}
        else:
            entity['scope'] = {'project': {'id': attribs['project_id']}}

        if attribs.get('user_id'):
            entity['user'] = {'id': attribs['user_id']}

            if attribs.get('group_id'):
                entity['links']['membership'] = '/groups/{}/users/{}'.format(
                    attribs['group_id'], attribs['user_id']
                )
        else:
            entity['group'] = {'id': attribs['group_id']}

        entity['role'] = {'id': attribs['role_id']}

        if attribs.get('inherited_to_projects'):
            entity['scope']['OS-INHERIT:inherited_to'] = 'projects'

        if prior_role_link:
            entity['links']['prior_role'] = prior_role_link

        return entity

    def build_role_assignment_entity_include_names(
        self,
        domain_ref=None,
        role_ref=None,
        group_ref=None,
        user_ref=None,
        project_ref=None,
        inherited_assignment=None,
    ):
        """Build and return a role assignment entity with provided attributes.

        The expected attributes are: domain_ref or project_ref,
        user_ref or group_ref, role_ref and, optionally, inherited_to_projects.
        """
        entity = {'links': {}}
        attributes_for_links = {}
        if project_ref:
            dmn_name = PROVIDERS.resource_api.get_domain(
                project_ref['domain_id']
            )['name']

            entity['scope'] = {
                'project': {
                    'id': project_ref['id'],
                    'name': project_ref['name'],
                    'domain': {
                        'id': project_ref['domain_id'],
                        'name': dmn_name,
                    },
                }
            }
            attributes_for_links['project_id'] = project_ref['id']
        else:
            entity['scope'] = {
                'domain': {'id': domain_ref['id'], 'name': domain_ref['name']}
            }
            attributes_for_links['domain_id'] = domain_ref['id']
        if user_ref:
            dmn_name = PROVIDERS.resource_api.get_domain(
                user_ref['domain_id']
            )['name']
            entity['user'] = {
                'id': user_ref['id'],
                'name': user_ref['name'],
                'domain': {'id': user_ref['domain_id'], 'name': dmn_name},
            }
            attributes_for_links['user_id'] = user_ref['id']
        else:
            dmn_name = PROVIDERS.resource_api.get_domain(
                group_ref['domain_id']
            )['name']
            entity['group'] = {
                'id': group_ref['id'],
                'name': group_ref['name'],
                'domain': {'id': group_ref['domain_id'], 'name': dmn_name},
            }
            attributes_for_links['group_id'] = group_ref['id']

        if role_ref:
            entity['role'] = {'id': role_ref['id'], 'name': role_ref['name']}
            if role_ref['domain_id']:
                dmn_name = PROVIDERS.resource_api.get_domain(
                    role_ref['domain_id']
                )['name']
                entity['role']['domain'] = {
                    'id': role_ref['domain_id'],
                    'name': dmn_name,
                }
            attributes_for_links['role_id'] = role_ref['id']

        if inherited_assignment:
            entity['scope']['OS-INHERIT:inherited_to'] = 'projects'
            attributes_for_links['inherited_to_projects'] = True

        entity['links']['assignment'] = self.build_role_assignment_link(
            **attributes_for_links
        )

        return entity


class PaginationTestCaseBase(RestfulTestCase):
    """Base test for the resource pagination."""

    resource_name: ty.Optional[str] = None
    config_group: ty.Optional[str] = None

    def setUp(self):
        super().setUp()
        if not self.resource_name:
            self.skipTest("Not testing the base")
        if not self.config_group:
            self.skipTest("Not testing the base")

    @abc.abstractmethod
    def _create_resources(self, count: int):
        """Method creating given count of test resources

        This is an abstract method and must be implemented by every test
        class inheriting from this base class.
        """
        pass

    def test_list_requested_limit(self):
        """Test requesting certain page size"""
        count_resources: int = 6
        self._create_resources(count_resources)

        for expected_length in range(1, count_resources):
            response = self.get(
                f"/{self.resource_name}s?limit={expected_length}"
            )
            res_list = response.json_body[f"{self.resource_name}s"]
            res_links = response.json_body["links"]
            self.assertEqual(
                expected_length,
                len(res_list),
                "Requested count of entries returned",
            )
            self.assertEqual(
                f"http://localhost/v3/{self.resource_name}s?limit={expected_length}&marker={res_list[-1]['id']}",
                res_links.get("next"),
                "Next page link contains limit and marker",
            )

        # Try fetching with explicitly very high page size
        response = self.get(
            f"/{self.resource_name}s?limit={count_resources+100}"
        )
        res_list = response.json_body[f"{self.resource_name}s"]
        res_links = response.json_body["links"]
        self.assertGreaterEqual(
            len(res_list), count_resources, "All test resources are returned"
        )
        self.assertIsNone(res_links.get("next"), "Next page link is empty")

        response = self.get(f"/{self.resource_name}s?limit=1&sort_key=name")
        res_list = response.json_body[f"{self.resource_name}s"]
        res_links = response.json_body["links"]
        self.assertEqual(
            f"http://localhost/v3/{self.resource_name}s?limit=1&sort_key=name&marker={res_list[-1]['id']}",
            res_links.get("next"),
            "Next page link contains limit and marker as well as other passed keys",
        )

        self.config_fixture.config(list_limit=2)
        response = self.get(f"/{self.resource_name}s?limit=10")
        res_list = response.json_body[f"{self.resource_name}s"]
        self.assertGreaterEqual(
            len(response.json_body[f"{self.resource_name}s"]),
            2,
            "Requested limit higher then default wins",
        )

    def test_list_requested_limit_too_high(self):
        """Test passing limit exceeding the system max"""
        self.config_fixture.config(max_db_limit=4)

        count_resources: int = 6
        self._create_resources(count_resources)

        response = self.get(f"/{self.resource_name}s?limit={count_resources}")
        res_list = response.json_body[f"{self.resource_name}s"]
        res_links = response.json_body['links']
        self.assertEqual(
            4,
            len(res_list),
            "max_db_limit overrides requested count of entries",
        )
        self.assertEqual(
            f"http://localhost/v3/{self.resource_name}s?limit=4&marker={res_list[-1]['id']}",
            res_links.get("next"),
            "Next page link contains corrected limit and marker",
        )

    def test_list_default_limit(self):
        """Tests listing resources without limit set"""
        count_resources: int = 6
        self._create_resources(count_resources)

        self.config_fixture.config(list_limit=2)

        response = self.get(f'/{self.resource_name}s')
        res_list = response.json_body[f"{self.resource_name}s"]
        res_links = response.json_body['links']
        self.assertEqual(2, len(res_list), "default limit is applied")
        self.assertEqual(
            f"http://localhost/v3/{self.resource_name}s?marker={res_list[-1]['id']}&limit=2",
            res_links.get("next"),
            "Next page link contains corrected limit and marker",
        )

        self.config_fixture.config(group=self.config_group, list_limit=3)

        response = self.get(f'/{self.resource_name}s')
        res_list = response.json_body[f"{self.resource_name}s"]
        res_links = response.json_body['links']
        self.assertEqual(3, len(res_list), "default resource limit is applied")
        self.assertEqual(
            f"http://localhost/v3/{self.resource_name}s?marker={res_list[-1]['id']}&limit=3",
            res_links.get("next"),
            "Next page link contains corrected limit and marker",
        )

    def _consume_paginated_list(
        self, limit: ty.Optional[int] = None
    ) -> tuple[list[dict], int]:
        """Fetch all paginated resources"""
        found_resources: list = []
        pages: int = 0
        if limit:
            response = self.get(f"/{self.resource_name}s?limit={limit}")
        else:
            response = self.get(f"/{self.resource_name}s")
            pages = 1
        res_links = response.json_body['links']
        while "next" in res_links:
            # Put fetched resources into the found list
            found_resources.extend(
                response.json_body.get(f"{self.resource_name}s")
            )

            # Get the relative url of the next page (since testhelper only supports that)
            next_url = res_links.get("next", "")
            if next_url:
                next_rel_url = next_url[next_url.find("/v3") + 3 :]
                # fetch next page
                response = self.get(next_rel_url)
                res_links = response.json_body['links']
                pages += 1
            else:
                break
        return (found_resources, pages)

    def test_list_consume_all_resources(self):
        """Test paginating through all resources (simulate user SDK)"""
        count_resources: int = 50
        page_size: int = 5
        self._create_resources(count_resources)
        response = self.get(f"/{self.resource_name}s")
        current_count = len(response.json_body[f"{self.resource_name}s"])

        # Set pagination default at 5
        self.config_fixture.config(
            group=self.config_group, list_limit=page_size
        )

        (found_resources, pages) = self._consume_paginated_list()
        self.assertGreaterEqual(
            len(found_resources),
            count_resources,
            "Fetched at least all pre-created resources",
        )
        self.assertGreaterEqual(
            pages,
            current_count // page_size,
            "At least as many pages consumed as expected",
        )

        (found_resources, pages) = self._consume_paginated_list(limit=100)
        self.assertGreaterEqual(
            len(found_resources),
            count_resources,
            "Fetched at least all pre-created resources",
        )
        self.assertGreaterEqual(
            pages,
            current_count // 100,
            "At least as many pages consumed as expected",
        )



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import uuid

import stevedore

from keystone.api._shared import authentication
from keystone import auth
from keystone.auth.plugins import base
from keystone.auth.plugins import mapped
from keystone import exception
from keystone.tests import unit
from keystone.tests.unit.ksfixtures import auth_plugins

# for testing purposes only
METHOD_NAME = 'simple_challenge_response'
EXPECTED_RESPONSE = uuid.uuid4().hex
DEMO_USER_ID = uuid.uuid4().hex


class SimpleChallengeResponse(base.AuthMethodHandler):
    def authenticate(self, auth_payload):
        response_data = {}
        if 'response' in auth_payload:
            if auth_payload['response'] != EXPECTED_RESPONSE:
                raise exception.Unauthorized('Wrong answer')

            response_data['user_id'] = DEMO_USER_ID
            return base.AuthHandlerResponse(
                status=True, response_body=None, response_data=response_data
            )
        else:
            return base.AuthHandlerResponse(
                status=False,
                response_body={
                    "challenge": "What's the name of your high school?"
                },
                response_data=None,
            )


class TestAuthPlugin(unit.SQLDriverOverrides, unit.TestCase):
    def test_unsupported_auth_method(self):
        method_name = uuid.uuid4().hex
        auth_data = {'methods': [method_name]}
        auth_data[method_name] = {'test': 'test'}
        auth_data = {'identity': auth_data}
        self.assertRaises(
            exception.AuthMethodNotSupported,
            auth.core.AuthInfo.create,
            auth_data,
        )

    @mock.patch.object(auth.core, '_get_auth_driver_manager')
    def test_addition_auth_steps(self, stevedore_mock):
        simple_challenge_plugin = SimpleChallengeResponse()
        extension = stevedore.extension.Extension(
            name='simple_challenge',
            entry_point=None,
            plugin=None,
            obj=simple_challenge_plugin,
        )
        test_manager = stevedore.DriverManager.make_test_instance(extension)
        stevedore_mock.return_value = test_manager

        self.useFixture(
            auth_plugins.ConfigAuthPlugins(
                self.config_fixture, methods=[METHOD_NAME]
            )
        )
        self.useFixture(auth_plugins.LoadAuthPlugins(METHOD_NAME))

        auth_data = {'methods': [METHOD_NAME]}
        auth_data[METHOD_NAME] = {'test': 'test'}
        auth_data = {'identity': auth_data}
        auth_info = auth.core.AuthInfo.create(auth_data)
        auth_context = auth.core.AuthContext(method_names=[])
        try:
            with self.make_request():
                authentication.authenticate(auth_info, auth_context)
        except exception.AdditionalAuthRequired as e:
            self.assertIn('methods', e.authentication)
            self.assertIn(METHOD_NAME, e.authentication['methods'])
            self.assertIn(METHOD_NAME, e.authentication)
            self.assertIn('challenge', e.authentication[METHOD_NAME])

        # test correct response
        auth_data = {'methods': [METHOD_NAME]}
        auth_data[METHOD_NAME] = {'response': EXPECTED_RESPONSE}
        auth_data = {'identity': auth_data}
        auth_info = auth.core.AuthInfo.create(auth_data)
        auth_context = auth.core.AuthContext(method_names=[])
        with self.make_request():
            authentication.authenticate(auth_info, auth_context)
        self.assertEqual(DEMO_USER_ID, auth_context['user_id'])

        # test incorrect response
        auth_data = {'methods': [METHOD_NAME]}
        auth_data[METHOD_NAME] = {'response': uuid.uuid4().hex}
        auth_data = {'identity': auth_data}
        auth_info = auth.core.AuthInfo.create(auth_data)
        auth_context = auth.core.AuthContext(method_names=[])
        with self.make_request():
            self.assertRaises(
                exception.Unauthorized,
                authentication.authenticate,
                auth_info,
                auth_context,
            )

    def test_duplicate_method(self):
        # Having the same method twice doesn't cause load_auth_methods to fail.
        self.useFixture(
            auth_plugins.ConfigAuthPlugins(
                self.config_fixture, ['external', 'external']
            )
        )
        auth.core.load_auth_methods()
        self.assertIn('external', auth.core.AUTH_METHODS)


class TestAuthPluginDynamicOptions(TestAuthPlugin):
    def config_overrides(self):
        super().config_overrides()
        # Clear the override for the [auth] ``methods`` option so it is
        # possible to load the options from the config file.
        self.config_fixture.conf.clear_override('methods', group='auth')

    def config_files(self):
        config_files = super().config_files()
        config_files.append(unit.dirs.tests_conf('test_auth_plugin.conf'))
        return config_files


class TestMapped(unit.TestCase):
    def config_files(self):
        config_files = super().config_files()
        config_files.append(unit.dirs.tests_conf('test_auth_plugin.conf'))
        return config_files

    def _test_mapped_invocation_with_method_name(self, method_name):
        with mock.patch.object(
            auth.plugins.mapped.Mapped, 'authenticate', return_value=None
        ) as authenticate:
            auth_data = {
                'identity': {
                    'methods': [method_name],
                    method_name: {'protocol': method_name},
                }
            }
            auth_info = auth.core.AuthInfo.create(auth_data)
            auth_context = auth.core.AuthContext(
                method_names=[], user_id=uuid.uuid4().hex
            )
            with self.make_request():
                authentication.authenticate(auth_info, auth_context)
            # make sure Mapped plugin got invoked with the correct payload
            ((auth_payload,), kwargs) = authenticate.call_args
            self.assertEqual(method_name, auth_payload['protocol'])

    def test_mapped_with_remote_user(self):
        method_name = 'saml2'
        auth_data = {'methods': [method_name]}
        # put the method name in the payload so its easier to correlate
        # method name with payload
        auth_data[method_name] = {'protocol': method_name}
        auth_data = {'identity': auth_data}

        auth_context = auth.core.AuthContext(
            method_names=[], user_id=uuid.uuid4().hex
        )

        self.useFixture(auth_plugins.LoadAuthPlugins(method_name))

        with mock.patch.object(
            auth.plugins.mapped.Mapped, 'authenticate', return_value=None
        ) as authenticate:
            auth_info = auth.core.AuthInfo.create(auth_data)
            with self.make_request(environ={'REMOTE_USER': 'foo@idp.com'}):
                authentication.authenticate(auth_info, auth_context)
            # make sure Mapped plugin got invoked with the correct payload
            ((auth_payload,), kwargs) = authenticate.call_args
            self.assertEqual(method_name, auth_payload['protocol'])

    @mock.patch('keystone.auth.plugins.mapped.PROVIDERS')
    def test_mapped_without_identity_provider_or_protocol(
        self, mock_providers
    ):
        mock_providers.resource_api = mock.Mock()
        mock_providers.federation_api = mock.Mock()
        mock_providers.identity_api = mock.Mock()
        mock_providers.assignment_api = mock.Mock()
        mock_providers.role_api = mock.Mock()

        test_mapped = mapped.Mapped()

        auth_payload = {'identity_provider': 'test_provider'}
        with self.make_request():
            self.assertRaises(
                exception.ValidationError,
                test_mapped.authenticate,
                auth_payload,
            )

        auth_payload = {'protocol': 'saml2'}
        with self.make_request():
            self.assertRaises(
                exception.ValidationError,
                test_mapped.authenticate,
                auth_payload,
            )

    def test_supporting_multiple_methods(self):
        method_names = ('saml2', 'openid', 'x509', 'mapped')
        self.useFixture(auth_plugins.LoadAuthPlugins(*method_names))
        for method_name in method_names:
            self._test_mapped_invocation_with_method_name(method_name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from keystone.common.policies import base

auth_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_auth_catalog',
        check_str='',
        description='Get service catalog.',
        operations=[
            {'path': '/v3/auth/catalog', 'method': 'GET'},
            {'path': '/v3/auth/catalog', 'method': 'HEAD'},
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_auth_projects',
        check_str='',
        description=(
            'List all projects a user has access to via role assignments.'
        ),
        operations=[
            {'path': '/v3/auth/projects', 'method': 'GET'},
            {'path': '/v3/auth/projects', 'method': 'HEAD'},
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_auth_domains',
        check_str='',
        description=(
            'List all domains a user has access to via role assignments.'
        ),
        operations=[
            {'path': '/v3/auth/domains', 'method': 'GET'},
            {'path': '/v3/auth/domains', 'method': 'HEAD'},
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_auth_system',
        check_str='',
        description='List systems a user has access to via role assignments.',
        operations=[
            {'path': '/v3/auth/system', 'method': 'GET'},
            {'path': '/v3/auth/system', 'method': 'HEAD'},
        ],
    ),
]


def list_rules():
    return auth_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Main entry point into the Revoke service."""

from keystone.common import cache
from keystone.common import manager
import keystone.conf
from keystone import exception
from keystone.i18n import _
from keystone.models import revoke_model
from keystone import notifications

CONF = keystone.conf.CONF

# This builds a discrete cache region dedicated to revoke events. The API can
# return a filtered list based upon last fetchtime. This is deprecated but
# must be maintained.
REVOKE_REGION = cache.create_region(name='revoke')
MEMOIZE = cache.get_memoization_decorator(group='revoke', region=REVOKE_REGION)


class Manager(manager.Manager):
    """Default pivot point for the Revoke backend.

    Performs common logic for recording revocations.

    See :mod:`keystone.common.manager.Manager` for more details on
    how this dynamically calls the backend.

    """

    driver_namespace = 'keystone.revoke'
    _provides_api = 'revoke_api'

    def __init__(self):
        super().__init__(CONF.revoke.driver)
        self._register_listeners()
        self.model = revoke_model

    @MEMOIZE
    def _list_events(self, last_fetch):
        return self.driver.list_events(last_fetch)

    def list_events(self, last_fetch=None):
        return self._list_events(last_fetch)

    def _user_callback(self, service, resource_type, operation, payload):
        self.revoke_by_user(payload['resource_info'])

    def _project_callback(self, service, resource_type, operation, payload):
        self.revoke(
            revoke_model.RevokeEvent(project_id=payload['resource_info'])
        )

    def _trust_callback(self, service, resource_type, operation, payload):
        self.revoke(
            revoke_model.RevokeEvent(trust_id=payload['resource_info'])
        )

    def _consumer_callback(self, service, resource_type, operation, payload):
        self.revoke(
            revoke_model.RevokeEvent(consumer_id=payload['resource_info'])
        )

    def _register_listeners(self):
        callbacks = {
            notifications.ACTIONS.deleted: [
                ['OS-TRUST:trust', self._trust_callback],
                ['OS-OAUTH1:consumer', self._consumer_callback],
                ['user', self._user_callback],
                ['project', self._project_callback],
            ],
            notifications.ACTIONS.disabled: [['user', self._user_callback]],
            notifications.ACTIONS.internal: [
                [
                    notifications.PERSIST_REVOCATION_EVENT_FOR_USER,
                    self._user_callback,
                ]
            ],
        }

        for event, cb_info in callbacks.items():
            for resource_type, callback_fns in cb_info:
                notifications.register_event_callback(
                    event, resource_type, callback_fns
                )

    def revoke_by_user(self, user_id):
        return self.revoke(revoke_model.RevokeEvent(user_id=user_id))

    def _assert_not_domain_and_project_scoped(
        self, domain_id=None, project_id=None
    ):
        if domain_id is not None and project_id is not None:
            msg = _(
                'The revoke call must not have both domain_id and '
                'project_id. This is a bug in the Keystone server. The '
                'current request is aborted.'
            )
            raise exception.UnexpectedError(exception=msg)

    def revoke_by_audit_id(self, audit_id):
        self.revoke(revoke_model.RevokeEvent(audit_id=audit_id))

    def revoke_by_audit_chain_id(
        self, audit_chain_id, project_id=None, domain_id=None
    ):
        self._assert_not_domain_and_project_scoped(
            domain_id=domain_id, project_id=project_id
        )

        self.revoke(
            revoke_model.RevokeEvent(
                audit_chain_id=audit_chain_id,
                domain_id=domain_id,
                project_id=project_id,
            )
        )

    def check_token(self, token):
        """Check the values from a token against the revocation list.

        :param token: dictionary of values from a token, normalized for
                             differences between v2 and v3. The checked values
                             are a subset of the attributes of model.TokenEvent

        :raises keystone.exception.TokenNotFound: If the token is invalid.

        """
        if revoke_model.is_revoked(
            self.driver.list_events(token=token), token
        ):
            raise exception.TokenNotFound(_('Failed to validate token'))

    def revoke(self, event):
        self.driver.revoke(event)
        REVOKE_REGION.invalidate()



# Copyright 2022 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import flask

from keystone.server import flask as ks_flask


def build_response(error):
    response = flask.make_response(
        (
            {
                'error': error.error_title,
                'error_description': error.message_format,
            },
            f"{error.code} {error.title}",
        )
    )

    if error.code == 401:
        response.headers['WWW-Authenticate'] = (
            f'Keystone uri="{ks_flask.base_url()}"'
        )
    return response
